# -*- coding: utf-8 -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2014-2015, Isis Lovecruft
#             (c) 2014-2015, The Tor Project, Inc.
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""Unittests for :mod:`bridgedb.https.server`."""

from __future__ import print_function

import logging
import os
import shutil

import ipaddr

from BeautifulSoup import BeautifulSoup

from twisted.internet import reactor
from twisted.internet import task
from twisted.trial import unittest
from twisted.web.resource import Resource
from twisted.web.test import requesthelper

from bridgedb.https import server
from bridgedb.schedule import ScheduledInterval

from .https_helpers import _createConfig
from .https_helpers import DummyRequest
from .https_helpers import DummyHTTPSDistributor
from .util import DummyBridge
from .util import DummyMaliciousBridge


# For additional logger output for debugging, comment out the following:
logging.disable(50)
# and then uncomment the following line:
#server.logging.getLogger().setLevel(10)


class SetFQDNTests(unittest.TestCase):
    """Tests for :func:`bridgedb.https.server.setFQDN` and
    :func:`bridgedb.https.server.setFQDN`.
    """

    def setUp(self):
        self.originalFQDN = server.SERVER_PUBLIC_FQDN

    def tearDown(self):
        server.SERVER_PUBLIC_FQDN = self.originalFQDN

    def test_setFQDN_https(self):
        """Calling ``server.setFQDN([…], https=True)`` should prepend
        ``"https://"`` to the module :data:`server.SERVER_PUBLIC_FQDN`
        variable.
        """
        server.setFQDN('example.com', https=True)
        self.assertEqual(server.SERVER_PUBLIC_FQDN, "https://example.com")

    def test_setFQDN_http(self):
        """Calling ``server.setFQDN([…], https=False)`` should not prepend
        anything at all to the module :data:`server.SERVER_PUBLIC_FQDN`
        variable.
        """
        server.setFQDN('example.com', https=False)
        self.assertEqual(server.SERVER_PUBLIC_FQDN, "example.com")


class GetClientIPTests(unittest.TestCase):
    """Tests for :func:`bridgedb.https.server.getClientIP`."""

    def createRequestWithIPs(self):
        """Set the IP address returned from ``request.getClientIP()`` to
        '3.3.3.3', and the IP address reported in the 'X-Forwarded-For' header
        to '2.2.2.2'.
        """
        request = DummyRequest([''])
        request.headers.update({'x-forwarded-for': '2.2.2.2'})
        # See :api:`twisted.test.requesthelper.DummyRequest.getClientIP`
        request.client = requesthelper.IPv4Address('TCP', '3.3.3.3', 443)
        request.method = b'GET'
        return request

    def test_getClientIP_XForwardedFor(self):
        """getClientIP() should return the IP address from the
        'X-Forwarded-For' header when ``useForwardedHeader=True``.
        """
        request = self.createRequestWithIPs()
        clientIP = server.getClientIP(request, useForwardedHeader=True)
        self.assertEqual(clientIP, '2.2.2.2')

    def test_getClientIP_XForwardedFor_bad_ip(self):
        """getClientIP() should return None if the IP address from the
        'X-Forwarded-For' header is bad/invalid and
        ``useForwardedHeader=True``.
        """
        request = self.createRequestWithIPs()
        request.headers.update({'x-forwarded-for': 'pineapple'})
        clientIP = server.getClientIP(request, useForwardedHeader=True)
        self.assertEqual(clientIP, None)

    def test_getClientIP_fromRequest(self):
        """getClientIP() should return the IP address from the request instance
        when ``useForwardedHeader=False``.
        """
        request = self.createRequestWithIPs()
        clientIP = server.getClientIP(request)
        self.assertEqual(clientIP, '3.3.3.3')


class ReplaceErrorPageTests(unittest.TestCase):
    """Tests for :func:`bridgedb.https.server.replaceErrorPage`."""

    def setUp(self):
        self.resource500 = server.resource500

    def tearDown(self):
        server.resource500 = self.resource500

    def test_replaceErrorPage(self):
        """``replaceErrorPage`` should return the error-500.html page."""
        request = DummyRequest([''])
        exc = Exception("vegan gümmibären")
        errorPage = server.replaceErrorPage(request, exc)
        self.assertSubstring("Bad News Bears", errorPage)
        self.assertNotSubstring("vegan gümmibären", errorPage)

    def test_replaceErrorPage_matches_resource500(self):
        """``replaceErrorPage`` should return the error-500.html page."""
        request = DummyRequest([''])
        exc = Exception("vegan gümmibären")
        errorPage = server.replaceErrorPage(request, exc)
        error500Page = server.resource500.render(request)
        self.assertEqual(errorPage, error500Page)

    def test_replaceErrorPage_no_resource500(self):
        """If ``server.resource500`` is missing/broken, then
        ``replaceErrorPage`` should return custom hardcoded HTML error text.
        """
        request = DummyRequest([''])
        exc = Exception("vegan gümmibären")
        server.resource500 = None
        errorPage = server.replaceErrorPage(request, exc)
        self.assertNotSubstring("Bad News Bears", errorPage)
        self.assertNotSubstring("vegan gümmibären", errorPage)
        self.assertSubstring("Sorry! Something went wrong with your request.",
                             errorPage)

class ErrorResourceTests(unittest.TestCase):
    """Tests for :class:`bridgedb.https.server.ErrorResource`."""

    def setUp(self):
        self.request = DummyRequest([''])

    def test_resource404(self):
        """``server.resource404`` should display the error-404.html page."""
        page = server.resource404.render(self.request)
        self.assertSubstring('We dug around for the page you requested', page)

    def test_resource500(self):
        """``server.resource500`` should display the error-500.html page."""
        page = server.resource500.render(self.request)
        self.assertSubstring('Bad News Bears', page)

    def test_maintenance(self):
        """``server.maintenance`` should display the error-503.html page."""
        page = server.maintenance.render(self.request)
        self.assertSubstring('Under Maintenance', page)


class CustomErrorHandlingResourceTests(unittest.TestCase):
    """Tests for :class:`bridgedb.https.server.CustomErrorHandlingResource`."""

    def test_getChild(self):
        """``CustomErrorHandlingResource.getChild`` should return a rendered
        copy of ``server.resource404``.
        """
        request = DummyRequest([''])
        resource = server.CustomErrorHandlingResource()
        self.assertEqual(server.resource404, resource.getChild('foobar', request))


class CSPResourceTests(unittest.TestCase):
    """Tests for :class:`bridgedb.HTTPServer.CSPResource`."""

    def setUp(self):
        self.pagename = b'foo.html'
        self.request = DummyRequest([self.pagename])
        self.request.method = b'GET'

        server.setFQDN('bridges.torproject.org')

    def test_CSPResource_setCSPHeader(self):
        """Setting the CSP header on a request should work out just peachy,
        like no errors or other bad stuff happening.
        """
        resource = server.CSPResource()
        resource.setCSPHeader(self.request)

    def test_render_POST_ascii(self):
        """Calling ``CSPResource.render_POST()`` should log whatever stuff was
        sent in the body of the POST request.
        """
        self.request.method = b'POST'
        self.request.writeContent('lah dee dah')
        self.request.client = requesthelper.IPv4Address('TCP', '3.3.3.3', 443)

        resource = server.CSPResource()
        page = resource.render_POST(self.request)

        self.assertIn('<html>', str(page))

    def test_render_POST_no_client_IP(self):
        """Calling ``CSPResource.render_POST()`` should log whatever stuff was
        sent in the body of the POST request, regardless of whether we were
        able to determine the client's IP address.
        """
        self.request.method = b'POST'
        self.request.writeContent('lah dee dah')

        resource = server.CSPResource()
        page = resource.render_POST(self.request)

        self.assertIn('<html>', str(page))

    def test_render_POST_unicode(self):
        """Calling ``CSPResource.render_POST()`` should log whatever stuff was
        sent in the body of the POST request, even if it's unicode.
        """
        self.request.method = b'POST'
        self.request.writeContent(
            ('南京大屠杀是中国抗日战争初期侵华日军在中华民国首都南京犯下的'
             '大規模屠殺、強姦以及纵火、抢劫等战争罪行与反人类罪行。'))
        self.request.client = requesthelper.IPv4Address('TCP', '3.3.3.3', 443)

        resource = server.CSPResource()
        page = resource.render_POST(self.request)

        self.assertIn('<html>', str(page))

    def test_render_POST_weird_request(self):
        """Calling ``CSPResource.render_POST()`` without a strange content
        object which doesn't have a ``content`` attribute should trigger the
        ``except Exception`` clause.
        """
        self.request.method = b'GET'
        del self.request.content

        resource = server.CSPResource()
        page = resource.render_POST(self.request)

        self.assertIn('<html>', str(page))


class IndexResourceTests(unittest.TestCase):
    """Test for :class:`bridgedb.https.server.IndexResource`."""

    def setUp(self):
        self.pagename = ''
        self.indexResource = server.IndexResource()
        self.root = Resource()
        self.root.putChild(self.pagename, self.indexResource)

    def test_IndexResource_render_GET(self):
        """renderGet() should return the index page."""
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        page = self.indexResource.render_GET(request)
        self.assertSubstring("add the bridges to Tor Browser", page)

    def test_IndexResource_render_GET_lang_ta(self):
        """renderGet() with ?lang=ta should return the index page in Tamil."""
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.addArg('lang', 'ta')
        page = self.indexResource.render_GET(request)
        self.assertSubstring("bridge-களை Tor Browser-உள்", page)


class HowtoResourceTests(unittest.TestCase):
    """Test for :class:`bridgedb.https.server.HowtoResource`."""

    def setUp(self):
        self.pagename = 'howto.html'
        self.howtoResource = server.HowtoResource()
        self.root = Resource()
        self.root.putChild(self.pagename, self.howtoResource)

    def test_HowtoResource_render_GET(self):
        """renderGet() should return the howto page."""
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        page = self.howtoResource.render_GET(request)
        self.assertSubstring("the wizard", page)

    def test_HowtoResource_render_GET_lang_ru(self):
        """renderGet() with ?lang=ru should return the howto page in Russian."""
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.addArg('lang', 'ru')
        page = self.howtoResource.render_GET(request)
        self.assertSubstring("следуйте инструкциям установщика", page)


class CaptchaProtectedResourceTests(unittest.TestCase):
    """Tests for :class:`bridgedb.https.server.CaptchaProtectedResource`."""

    def setUp(self):
        self.dist = None
        self.sched = None
        self.pagename = b'bridges.html'
        self.root = Resource()
        self.protectedResource = server.BridgesResource(self.dist, self.sched)
        self.captchaResource = server.CaptchaProtectedResource(
            useForwardedHeader=True, protectedResource=self.protectedResource)
        self.root.putChild(self.pagename, self.captchaResource)

    def test_render_GET_noCaptcha(self):
        """render_GET() should return a page without a CAPTCHA, which has the
        image alt text.
        """
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        page = self.captchaResource.render_GET(request)
        self.assertSubstring(
            "Your browser is not displaying images properly", page)

    def test_render_GET_missingTemplate(self):
        """render_GET() with a missing template should raise an error and
        return the result of replaceErrorPage().
        """
        oldLookup = server.lookup
        try:
            server.lookup = None
            request = DummyRequest([self.pagename])
            request.method = b'GET'
            page = self.captchaResource.render_GET(request)
            errorPage = server.replaceErrorPage(request, Exception('kablam'))
            self.assertEqual(page, errorPage)
        finally:
            server.lookup = oldLookup

    def createRequestWithIPs(self):
        """Set the IP address returned from ``request.getClientIP()`` to
        '3.3.3.3', and the IP address reported in the 'X-Forwarded-For' header
        to '2.2.2.2'.
        """
        request = DummyRequest([self.pagename])
        # Since we do not set ``request.getClientIP`` here like we do in some
        # of the other unittests, an exception would be raised here if
        # ``getBridgesForRequest()`` is unable to get the IP address from this
        # 'X-Forwarded-For' header (because ``ip`` would get set to ``None``).
        request.headers.update({'x-forwarded-for': '2.2.2.2'})
        # See :api:`twisted.test.requesthelper.DummyRequest.getClientIP`
        request.client = requesthelper.IPv4Address('TCP', '3.3.3.3', 443)
        request.method = b'GET'
        return request

    def test_getClientIP_XForwardedFor(self):
        """CaptchaProtectedResource.getClientIP() should return the IP address
        from the 'X-Forwarded-For' header when ``useForwardedHeader=True``.
        """
        self.captchaResource.useForwardedHeader = True
        request = self.createRequestWithIPs()
        clientIP = self.captchaResource.getClientIP(request)
        self.assertEqual(clientIP, '2.2.2.2')

    def test_getClientIP_fromRequest(self):
        """CaptchaProtectedResource.getClientIP() should return the IP address
        from the request instance when ``useForwardedHeader=False``.
        """
        self.captchaResource.useForwardedHeader = False
        request = self.createRequestWithIPs()
        clientIP = self.captchaResource.getClientIP(request)
        self.assertEqual(clientIP, '3.3.3.3')

    def test_render_POST(self):
        """render_POST() with a wrong 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        request = DummyRequest([self.pagename])
        request.method = b'POST'
        page = self.captchaResource.render_POST(request)
        self.assertEqual(BeautifulSoup(page).find('meta')['http-equiv'],
                         'refresh')


class GimpCaptchaProtectedResourceTests(unittest.TestCase):
    """Tests for :mod:`bridgedb.https.server.GimpCaptchaProtectedResource`."""

    def setUp(self):
        """Create a :class:`server.BridgesResource` and protect it with
        a :class:`GimpCaptchaProtectedResource`.
        """
        # Create our cached CAPTCHA directory:
        self.captchaDir = 'captchas'
        if not os.path.isdir(self.captchaDir):
            os.makedirs(self.captchaDir)

        # Set up our resources to fake a minimal HTTP(S) server:
        self.pagename = b'captcha.html'
        self.root = Resource()
        # (None, None) is the (distributor, scheduleInterval):
        self.protectedResource = server.BridgesResource(None, None)
        self.captchaResource = server.GimpCaptchaProtectedResource(
            secretKey='42',
            publicKey='23',
            hmacKey='abcdefghijklmnopqrstuvwxyz012345',
            captchaDir='captchas',
            useForwardedHeader=True,
            protectedResource=self.protectedResource)

        self.root.putChild(self.pagename, self.captchaResource)

        # Set up the basic parts of our faked request:
        self.request = DummyRequest([self.pagename])

    def tearDown(self):
        """Delete the cached CAPTCHA directory if it still exists."""
        if os.path.isdir(self.captchaDir):
            shutil.rmtree(self.captchaDir)

    def test_extractClientSolution(self):
        """A (challenge, sollution) pair extracted from a request resulting
        from a POST should have the same unmodified (challenge, sollution) as
        the client originally POSTed.
        """
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        response = self.captchaResource.extractClientSolution(self.request)
        (challenge, response) = response
        self.assertEqual(challenge, expectedChallenge)
        self.assertEqual(response, expectedResponse)

    def test_checkSolution(self):
        """checkSolution() should return False is the solution is invalid."""
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        valid = self.captchaResource.checkSolution(self.request)
        self.assertFalse(valid)

    def test_getCaptchaImage(self):
        """Retrieving a (captcha, challenge) pair with an empty captchaDir
        should return None for both of the (captcha, challenge) strings.
        """
        self.request.method = b'GET'
        response = self.captchaResource.getCaptchaImage(self.request)
        (image, challenge) = response
        # Because we created the directory, there weren't any CAPTCHAs to
        # retrieve from it:
        self.assertIs(image, None)
        self.assertIs(challenge, None)

    def test_getCaptchaImage_noCaptchaDir(self):
        """Retrieving a (captcha, challenge) with an missing captchaDir should
        raise a bridgedb.captcha.GimpCaptchaError.
        """
        shutil.rmtree(self.captchaDir)
        self.request.method = b'GET'
        self.assertRaises(server.captcha.GimpCaptchaError,
                          self.captchaResource.getCaptchaImage, self.request)

    def test_render_GET_missingTemplate(self):
        """render_GET() with a missing template should raise an error and
        return the result of replaceErrorPage().
        """
        oldLookup = server.lookup
        try:
            server.lookup = None
            self.request.method = b'GET'
            page = self.captchaResource.render_GET(self.request)
            errorPage = server.replaceErrorPage(self.request, Exception('kablam'))
            self.assertEqual(page, errorPage)
        finally:
            server.lookup = oldLookup

    def test_render_POST_blankFields(self):
        """render_POST() with a blank 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', '')
        self.request.addArg('captcha_response_field', '')

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(BeautifulSoup(page).find('meta')['http-equiv'],
                         'refresh')

    def test_render_POST_wrongSolution(self):
        """render_POST() with a wrong 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(BeautifulSoup(page).find('meta')['http-equiv'],
                         'refresh')


class ReCaptchaProtectedResourceTests(unittest.TestCase):
    """Tests for :mod:`bridgedb.https.server.ReCaptchaProtectedResource`."""

    def setUp(self):
        """Create a :class:`server.BridgesResource` and protect it with
        a :class:`ReCaptchaProtectedResource`.
        """
        self.timeout = 10.0  # Can't take longer than that, right?
        # Set up our resources to fake a minimal HTTP(S) server:
        self.pagename = b'captcha.html'
        self.root = Resource()
        # (None, None) is the (distributor, scheduleInterval):
        self.protectedResource = server.BridgesResource(None, None)
        self.captchaResource = server.ReCaptchaProtectedResource(
            publicKey='23',
            secretKey='42',
            remoteIP='111.111.111.111',
            useForwardedHeader=True,
            protectedResource=self.protectedResource)

        self.root.putChild(self.pagename, self.captchaResource)

        # Set up the basic parts of our faked request:
        self.request = DummyRequest([self.pagename])

    def tearDown(self):
        """Cleanup method for removing timed out connections on the reactor.

        This seems to be the solution for the dirty reactor due to
        ``DelayedCall``s which is mentioned at the beginning of this
        file. There doesn't seem to be any documentation anywhere which
        proposes this solution, although this seems to solve the problem.
        """
        for delay in reactor.getDelayedCalls():
            try:
                delay.cancel()
            except (AlreadyCalled, AlreadyCancelled):
                pass

    def test_renderDeferred_invalid(self):
        """:meth:`_renderDeferred` should redirect a ``Request`` (after the
        CAPTCHA was NOT xsuccessfully solved) which results from a
        ``Deferred``'s callback.
        """
        self.request.method = b'POST'

        def testCB(request):
            """Check the ``Request`` returned from ``_renderDeferred``."""
            self.assertIsInstance(request, DummyRequest)
            soup = BeautifulSoup(b''.join(request.written)).find('meta')['http-equiv']
            self.assertEqual(soup, 'refresh')

        d = task.deferLater(reactor, 0, lambda x: x, (False, self.request))
        d.addCallback(self.captchaResource._renderDeferred)
        d.addCallback(testCB)
        return d

    def test_renderDeferred_valid(self):
        """:meth:`_renderDeferred` should correctly render a ``Request`` (after
        the CAPTCHA has been successfully solved) which results from a
        ``Deferred``'s callback.
        """
        self.request.method = b'POST'

        def testCB(request):
            """Check the ``Request`` returned from ``_renderDeferred``."""
            self.assertIsInstance(request, DummyRequest)
            html = b''.join(request.written)
            self.assertSubstring('Uh oh, spaghettios!', html)

        d = task.deferLater(reactor, 0, lambda x: x, (True, self.request))
        d.addCallback(self.captchaResource._renderDeferred)
        d.addCallback(testCB)
        return d

    def test_renderDeferred_nontuple(self):
        """:meth:`_renderDeferred` should correctly render a ``Request`` (after
        the CAPTCHA has been successfully solved) which results from a
        ``Deferred``'s callback.
        """
        self.request.method = b'POST'

        def testCB(request):
            """Check the ``Request`` returned from ``_renderDeferred``."""
            self.assertIs(request, None)

        d = task.deferLater(reactor, 0, lambda x: x, (self.request))
        d.addCallback(self.captchaResource._renderDeferred)
        d.addCallback(testCB)
        return d

    def test_checkSolution_blankFields(self):
        """:meth:`server.ReCaptchaProtectedResource.checkSolution` should
        return a redirect if is the solution field is blank.
        """
        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', '')
        self.request.addArg('captcha_response_field', '')

        self.assertEqual((False, self.request),
                         self.successResultOf(
                             self.captchaResource.checkSolution(self.request)))

    def test_getRemoteIP_useRandomIP(self):
        """Check that removing our remoteip setting produces a random IP."""
        self.captchaResource.remoteIP = None
        ip = self.captchaResource.getRemoteIP()
        realishIP = ipaddr.IPv4Address(ip).compressed
        self.assertTrue(realishIP)
        self.assertNotEquals(realishIP, '111.111.111.111')

    def test_getRemoteIP_useConfiguredIP(self):
        """Check that our remoteip setting is used if configured."""
        ip = self.captchaResource.getRemoteIP()
        realishIP = ipaddr.IPv4Address(ip).compressed
        self.assertTrue(realishIP)
        self.assertEquals(realishIP, '111.111.111.111')

    def test_render_GET_missingTemplate(self):
        """render_GET() with a missing template should raise an error and
        return the result of replaceErrorPage().
        """
        oldLookup = server.lookup
        try:
            server.lookup = None
            self.request.method = b'GET'
            page = self.captchaResource.render_GET(self.request)
            errorPage = server.replaceErrorPage(self.request, Exception('kablam'))
            self.assertEqual(page, errorPage)
        finally:
            server.lookup = oldLookup

    def test_render_POST_blankFields(self):
        """render_POST() with a blank 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', '')
        self.request.addArg('captcha_response_field', '')

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(page, server.NOT_DONE_YET)

    def test_render_POST_wrongSolution(self):
        """render_POST() with a wrong 'captcha_response_field' should return
        a redirect to the CaptchaProtectedResource page.
        """
        expectedChallenge = '23232323232323232323'
        expectedResponse = 'awefawefaefawefaewf'

        self.request.method = b'POST'
        self.request.addArg('captcha_challenge_field', expectedChallenge)
        self.request.addArg('captcha_response_field', expectedResponse)

        page = self.captchaResource.render_POST(self.request)
        self.assertEqual(page, server.NOT_DONE_YET)


class BridgesResourceTests(unittest.TestCase):
    """Tests for :class:`https.server.BridgesResource`."""

    def setUp(self):
        """Set up our resources to fake a minimal HTTP(S) server."""
        self.pagename = b'bridges.html'
        self.root = Resource()

        self.dist = DummyHTTPSDistributor()
        self.sched = ScheduledInterval(1, 'hour')
        self.nBridgesPerRequest = 2

    def useBenignBridges(self):
        self.dist._bridge_class = DummyBridge
        self.bridgesResource = server.BridgesResource(
            self.dist, self.sched, N=self.nBridgesPerRequest,
            includeFingerprints=True)
        self.root.putChild(self.pagename, self.bridgesResource)

    def useMaliciousBridges(self):
        self.dist._bridge_class = DummyMaliciousBridge
        self.bridgesResource = server.BridgesResource(
            self.dist, self.sched, N=self.nBridgesPerRequest,
            includeFingerprints=True)
        self.root.putChild(self.pagename, self.bridgesResource)

    def parseBridgesFromHTMLPage(self, page):
        """Utility to pull the bridge lines out of an HTML response page.

        :param str page: A rendered HTML page, as a string.
        :raises: Any error which might occur.
        :rtype: list
        :returns: A list of the bridge lines contained on the **page**.
        """
        # The bridge lines are contained in a <div class='bridges'> tag:
        soup = BeautifulSoup(page)
        well = soup.find('div', {'class': 'bridge-lines'})
        content = well.renderContents().strip()
        lines = content.splitlines()

        bridges = []
        for line in lines:
            bridgelines = line.split('<br />')
            for bridge in bridgelines:
                if bridge:  # It still could be an empty string at this point
                    bridges.append(bridge)

        return bridges

    def test_render_GET_malicious_newlines(self):
        """Test rendering a request when the some of the bridges returned have
        malicious (HTML, Javascript, etc., in their) PT arguments.
        """
        self.useMaliciousBridges()

        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.getClientIP = lambda: '1.1.1.1'

        page = self.bridgesResource.render(request)
        self.assertTrue(
            'bad=Bridge 6.6.6.6:6666 0123456789abcdef0123456789abcdef01234567' in str(page),
            "Newlines in bridge lines should be removed.")

    def test_render_GET_malicious_returnchar(self):
        """Test rendering a request when the some of the bridges returned have
        malicious (HTML, Javascript, etc., in their) PT arguments.
        """
        self.useMaliciousBridges()

        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.getClientIP = lambda: '1.1.1.1'

        page = self.bridgesResource.render(request)
        self.assertTrue(
            'eww=Bridge 1.2.3.4:1234' in str(page),
            "Return characters in bridge lines should be removed.")

    def test_render_GET_malicious_javascript(self):
        """Test rendering a request when the some of the bridges returned have
        malicious (HTML, Javascript, etc., in their) PT arguments.
        """
        self.useMaliciousBridges()

        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.getClientIP = lambda: '1.1.1.1'

        page = self.bridgesResource.render(request)
        self.assertTrue(
            "evil=&lt;script&gt;alert(&#39;fuuuu&#39;);&lt;/script&gt;" in str(page),
            ("The characters &, <, >, ', and \" in bridge lines should be "
             "replaced with their corresponding HTML special characters."))

    def test_renderAnswer_GET_textplain_malicious(self):
        """If the request format specifies 'plain', we should return content
        with mimetype 'text/plain' and ASCII control characters replaced.
        """
        self.useMaliciousBridges()

        request = DummyRequest([self.pagename])
        request.args.update({'format': ['plain']})
        request.getClientIP = lambda: '4.4.4.4'
        request.method = b'GET'

        page = self.bridgesResource.render(request)
        self.assertTrue("html" not in str(page))
        self.assertTrue(
            'eww=Bridge 1.2.3.4:1234' in str(page),
            "Return characters in bridge lines should be removed.")
        self.assertTrue(
            'bad=Bridge 6.6.6.6:6666' in str(page),
            "Newlines in bridge lines should be removed.")

    def test_render_GET_vanilla(self):
        """Test rendering a request for normal, vanilla bridges."""
        self.useBenignBridges()

        request = DummyRequest([self.pagename])
        request.method = b'GET'
        request.getClientIP = lambda: '1.1.1.1'

        page = self.bridgesResource.render(request)

        # The response should explain how to use the bridge lines:
        self.assertTrue("To enter bridges into Tor Browser" in str(page))

        for b in self.parseBridgesFromHTMLPage(page):
            # Check that each bridge line had the expected number of fields:
            fields = b.split(' ')
            self.assertEqual(len(fields), 2)

            # Check that the IP and port seem okay:
            ip, port = fields[0].rsplit(':')
            self.assertIsInstance(ipaddr.IPv4Address(ip), ipaddr.IPv4Address)
            self.assertIsInstance(int(port), int)
            self.assertGreater(int(port), 0)
            self.assertLessEqual(int(port), 65535)

    def test_render_GET_XForwardedFor(self):
        """The client's IP address should be obtainable from the
        'X-Forwarded-For' header in the request.
        """
        self.useBenignBridges()

        self.bridgesResource.useForwardedHeader = True
        request = DummyRequest([self.pagename])
        request.method = b'GET'
        # Since we do not set ``request.getClientIP`` here like we do in some
        # of the other unittests, an exception would be raised here if
        # ``getBridgesForRequest()`` is unable to get the IP address from this
        # 'X-Forwarded-For' header (because ``ip`` would get set to ``None``).
        request.headers.update({'x-forwarded-for': '2.2.2.2'})

        page = self.bridgesResource.render(request)
        self.bridgesResource.useForwardedHeader = False  # Reset it

        # The response should explain how to use the bridge lines:
        self.assertTrue("To enter bridges into Tor Browser" in str(page))

    def test_render_GET_RTLlang(self):
        """Test rendering a request for plain bridges in Arabic."""
        self.useBenignBridges()

        request = DummyRequest([b"bridges?transport=obfs3"])
        request.method = b'GET'
        request.getClientIP = lambda: '3.3.3.3'
        # For some strange reason, the 'Accept-Language' value *should not* be
        # a list, unlike all the other headers and args…
        request.headers.update({'accept-language': 'ar,en,en_US,'})

        page = self.bridgesResource.render(request)
        self.assertSubstring("rtl.css", page)
        self.assertSubstring(
            # "I need an alternative way to get bridges!"
            "أحتاج إلى وسيلة بديلة للحصول على bridges", page)

        for bridgeLine in self.parseBridgesFromHTMLPage(page):
            # Check that each bridge line had the expected number of fields:
            bridgeLine = bridgeLine.split(' ')
            self.assertEqual(len(bridgeLine), 2)

    def test_render_GET_RTLlang_obfs3(self):
        """Test rendering a request for obfs3 bridges in Farsi."""
        self.useBenignBridges()

        request = DummyRequest([b"bridges?transport=obfs3"])
        request.method = b'GET'
        request.getClientIP = lambda: '3.3.3.3'
        request.headers.update({'accept-language': 'fa,en,en_US,'})
        # We actually have to set the request args manually when using a
        # DummyRequest:
        request.args.update({'transport': ['obfs3']})

        page = self.bridgesResource.render(request)
        self.assertSubstring("rtl.css", page)
        self.assertSubstring(
            # "How to use the above bridge lines" (since there should be
            # bridges in this response, we don't tell them about alternative
            # mechanisms for getting bridges)
            "چگونگی از پل‌های خود استفاده کنید", page)

        for bridgeLine in self.parseBridgesFromHTMLPage(page):
            # Check that each bridge line had the expected number of fields:
            bridgeLine = bridgeLine.split(' ')
            self.assertEqual(len(bridgeLine), 3)
            self.assertEqual(bridgeLine[0], 'obfs3')

            # Check that the IP and port seem okay:
            ip, port = bridgeLine[1].rsplit(':')
            self.assertIsInstance(ipaddr.IPv4Address(ip), ipaddr.IPv4Address)
            self.assertIsInstance(int(port), int)
            self.assertGreater(int(port), 0)
            self.assertLessEqual(int(port), 65535)

    def test_renderAnswer_textplain(self):
        """If the request format specifies 'plain', we should return content
        with mimetype 'text/plain'.
        """
        self.useBenignBridges()

        request = DummyRequest([self.pagename])
        request.args.update({'format': ['plain']})
        request.getClientIP = lambda: '4.4.4.4'
        request.method = b'GET'

        page = self.bridgesResource.render(request)
        self.assertTrue("html" not in str(page))

        # We just need to strip and split it because it looks like:
        #
        #   94.235.85.233:9492 0d9d0547c3471cddc473f7288a6abfb54562dc06
        #   255.225.204.145:9511 1fb89d618b3a12afe3529fd072127ea08fb50466
        #
        # (Yes, there are two leading spaces at the beginning of each line)
        #
        bridgeLines = [line.strip() for line in page.strip().split('\n')]

        for bridgeLine in bridgeLines:
            bridgeLine = bridgeLine.split(' ')
            self.assertEqual(len(bridgeLine), 2)

            # Check that the IP and port seem okay:
            ip, port = bridgeLine[0].rsplit(':')
            self.assertIsInstance(ipaddr.IPv4Address(ip), ipaddr.IPv4Address)
            self.assertIsInstance(int(port), int)
            self.assertGreater(int(port), 0)
            self.assertLessEqual(int(port), 65535)

    def test_renderAnswer_textplain_error(self):
        """If we hit some error while returning bridge lines in text/plain
        format, then our custom plaintext error message (the hardcoded HTML in
        ``server.replaceErrorPage``) should be returned.
        """
        self.useBenignBridges()

        request = DummyRequest([self.pagename])
        request.args.update({'format': ['plain']})
        request.getClientIP = lambda: '4.4.4.4'
        request.method = b'GET'

        # We'll cause a TypeError here due to calling '\n'.join(None)
        page = self.bridgesResource.renderAnswer(request, bridgeLines=None)

        # We don't want the fancy version:
        self.assertNotSubstring("Bad News Bears", page)
        self.assertSubstring("Sorry! Something went wrong with your request.",
                             page)


class OptionsResourceTests(unittest.TestCase):
    """Tests for :class:`bridgedb.https.server.OptionsResource`."""

    def setUp(self):
        """Create a :class:`server.OptionsResource`."""
        # Set up our resources to fake a minimal HTTP(S) server:
        self.pagename = b'options.html'
        self.root = Resource()
        self.optionsResource = server.OptionsResource()
        self.root.putChild(self.pagename, self.optionsResource)

    def test_render_GET_RTLlang(self):
        """Test rendering a request for obfs3 bridges in Hebrew."""
        request = DummyRequest(["bridges?transport=obfs3"])
        request.method = b'GET'
        request.getClientIP = lambda: '3.3.3.3'
        request.headers.update({'accept-language': 'he'})
        # We actually have to set the request args manually when using a
        # DummyRequest:
        request.args.update({'transport': ['obfs2']})

        page = self.optionsResource.render(request)
        self.assertSubstring("rtl.css", page)
        self.assertSubstring("מהם גשרים?", page)


class HTTPSServerServiceTests(unittest.TestCase):
    """Unittests for :func:`bridgedb.email.server.addWebServer`."""

    def setUp(self):
        """Create a config and an HTTPSDistributor."""
        self.config = _createConfig()
        self.distributor = DummyHTTPSDistributor()

    def tearDown(self):
        """Cleanup method after each ``test_*`` method runs; removes timed out
        connections on the reactor and clears the :ivar:`transport`.

        Basically, kill all connections with fire.
        """
        for delay in reactor.getDelayedCalls():
            try:
                delay.cancel()
            except (AlreadyCalled, AlreadyCancelled):
                pass

        # FIXME: this is definitely not how we're supposed to do this, but it
        # kills the DirtyReactorAggregateErrors.
        reactor.disconnectAll()
        reactor.runUntilCurrent()

    def test_addWebServer_GIMP_CAPTCHA_ENABLED(self):
        """Call :func:`bridgedb.https.server.addWebServer` to test startup."""
        server.addWebServer(self.config, self.distributor)

    def test_addWebServer_RECAPTCHA_ENABLED(self):
        """Call :func:`bridgedb.https.server.addWebServer` to test startup."""
        config = self.config
        config.RECAPTCHA_ENABLED = True
        server.addWebServer(config, self.distributor)

    def test_addWebServer_no_captchas(self):
        """Call :func:`bridgedb.https.server.addWebServer` to test startup."""
        config = self.config
        config.GIMP_CAPTCHA_ENABLED = False
        server.addWebServer(config, self.distributor)

    def test_addWebServer_no_HTTPS_ROTATION_PERIOD(self):
        """Call :func:`bridgedb.https.server.addWebServer` to test startup."""
        config = self.config
        config.HTTPS_ROTATION_PERIOD = None
        server.addWebServer(config, self.distributor)

    def test_addWebServer_CSP_ENABLED_False(self):
        """Call :func:`bridgedb.https.server.addWebServer` with
        ``CSP_ENABLED=False`` to test startup.
        """
        config = self.config
        config.CSP_ENABLED = False
        server.addWebServer(config, self.distributor)

    def test_addWebServer_CSP_REPORT_ONLY_False(self):
        """Call :func:`bridgedb.https.server.addWebServer` with
        ``CSP_REPORT_ONLY=False`` to test startup.
        """
        config = self.config
        config.CSP_REPORT_ONLY = False
        server.addWebServer(config, self.distributor)

    def test_addWebServer_CSP_INCLUDE_SELF_False(self):
        """Call :func:`bridgedb.https.server.addWebServer` with
        ``CSP_INCLUDE_SELF=False`` to test startup.
        """
        config = self.config
        config.CSP_INCLUDE_SELF = False
        server.addWebServer(config, self.distributor)



# -*- coding: utf-8 -*-
#_____________________________________________________________________________
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2014-2015, The Tor Project, Inc.
#             (c) 2014-2015, Isis Lovecruft
# :license: see LICENSE for licensing information
#_____________________________________________________________________________

"""Unittests for :mod:`bridgedb.parse.options`."""


from __future__ import print_function

import os
import sys

from twisted.python.usage import UsageError
from twisted.trial import unittest

from bridgedb.parse import options


class ParseOptionsTests(unittest.TestCase):
    """Unittests for :mod:`bridgedb.parse.options`."""

    def setUp(self):
        """Replace the current sys.argv's for the run of this test, and
        redirect sys.stdout to os.devnull to prevent the options parser from
        printing the --help a bunch of times.
        """
        # Make sure a config file is in the current directory, or else the
        # argument parser will get angry and throw another SystemExit
        # exception.
        with open(os.path.join(os.getcwd(), 'bridgedb.conf'), 'a+') as fh:
            fh.write('\n')

        self.oldSysArgv = sys.argv
        self.oldStdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def tearDown(self):
        """Put the original sys.argv's back."""
        sys.stdout.close()  # Actually closes the FD we opened for /dev/null
        sys.argv = self.oldSysArgv
        sys.stdout = self.oldStdout
        self.oldSysArgv = None
        self.oldStdout = None

    def test_parse_options_parseOptions_with_invalid_options(self):
        """:func:`options.parseOptions` should raise SystemExit because
        the args 'somearg anotherarg' are invalid commands.
        """
        fakeSysArgv = ['somearg', 'anotherarg']
        sys.argv = fakeSysArgv
        self.assertRaises(SystemExit, options.parseOptions)

    def test_parse_options_parseOptions_with_valid_options(self):
        """:func:`options.parseOptions` should return a
        :class:`options.MainOptions` when given valid commandline arguments.
        """
        fakeSysArgv = ['bridgedb', 'mock', '-n', '-1']
        sys.argv = fakeSysArgv
        opts = options.parseOptions()
        self.assertIsInstance(opts, options.MainOptions)

    def test_parse_options_parseOptions_verbosity_quiet_quiet(self):
        """If we use `-q` twice on the commandline, ``opts['verbosity']``
        should equal ``10``.
        """
        fakeSysArgv = ['bridgedb', '-q', '-q', 'mock', '-n', '-1']
        sys.argv = fakeSysArgv
        opts = options.parseOptions()
        self.assertEqual(opts['verbosity'], 10)

    def test_parse_options_parseOptions_verbosity_verbose(self):
        """If we use `-v` once on the commandline, ``opts['verbosity']``
        should equal ``50``.
        """
        fakeSysArgv = ['bridgedb', '-v', '-v', 'mock', '-n', '-1']
        sys.argv = fakeSysArgv
        opts = options.parseOptions()
        self.assertEqual(opts['verbosity'], 50)

    def test_parse_options_parseOptions_rundir(self):
        """The automatic rundir should be our current directory."""
        fakeSysArgv = ['bridgedb', 'mock', '-n', '-1']
        sys.argv = fakeSysArgv
        opts = options.parseOptions()
        self.assertEqual(opts['rundir'], os.getcwd())

    def test_parse_options_parseOptions_version(self):
        """:func:`options.parseOptions` when given a `--version` argument on
        the commandline, should raise SystemExit (after printing some stuff,
        but we don't care what it prints).
        """
        fakeSysArgv = ['bridgedb', '--version']
        sys.argv = fakeSysArgv
        self.assertRaises(SystemExit, options.parseOptions)



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013-2015, Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-Clause BSD, see LICENSE for licensing information

"""Unittests for :mod:`bridgedb.crypto`."""

from __future__ import print_function
from __future__ import unicode_literals

import base64
import io
import logging
import math
import os
import shutil

import OpenSSL

from twisted import version as _twistedversion
from twisted.internet import defer
from twisted.python.versions import Version
from twisted.trial import unittest
from twisted.test.proto_helpers import StringTransport
from twisted.web.test import test_agent as txtagent

from bridgedb import crypto
from bridgedb import txrecaptcha
from bridgedb.persistent import Conf

from .util import fileCheckDecorator
from .email_helpers import _createConfig


logging.disable(50)

SEKRIT_KEY  = b'v\x16Xm\xfc\x1b}\x063\x85\xaa\xa5\xf9\xad\x18\xb2P\x93\xc6k\xf9'
SEKRIT_KEY += b'\x8bI\xd9\xb8xw\xf5\xec\x1b\x7f\xa8'


class DummyEndpoint(object):
    """An endpoint that uses a fake transport."""

    def connect(self, factory):
        """Returns a connection to a
        :api:`twisted.test.proto_helpers.StringTransport`.
        """
        protocol = factory.buildProtocol(None)
        protocol.makeConnection(StringTransport())
        return defer.succeed(protocol)


class GetKeyTests(unittest.TestCase):
    """Tests for :func:`bridgedb.crypto.getKey`."""

    def test_getKey_nokey(self):
        """Test retrieving the secret_key from an empty file."""
        filename = os.path.join(os.getcwd(), 'sekrit')
        key = crypto.getKey(filename)
        self.failUnlessIsInstance(key, basestring,
                                  "key isn't a string! type=%r" % type(key))

    def test_getKey_tmpfile(self):
        """Test retrieving the secret_key from a new tmpfile."""
        filename = self.mktemp()
        key = crypto.getKey(filename)
        self.failUnlessIsInstance(key, basestring,
                                  "key isn't a string! type=%r" % type(key))

    def test_getKey_keyexists(self):
        """Write the example key to a file and test reading it back."""
        filename = self.mktemp()
        with open(filename, 'wb') as fh:
            fh.write(SEKRIT_KEY)
            fh.flush()

        key = crypto.getKey(filename)
        self.failUnlessIsInstance(key, basestring,
                                  "key isn't a string! type=%r" % type(key))
        self.assertEqual(SEKRIT_KEY, key,
                         """The example key and the one read from file differ!
                         key (in hex): %s
                         SEKRIT_KEY (in hex): %s"""
                         % (key.encode('hex'), SEKRIT_KEY.encode('hex')))


class InitializeGnuPGTests(unittest.TestCase):
    """Unittests for :func:`bridgedb.crypto.initializeGnupG`."""

    def _moveGnuPGHomedir(self):
        """Move the .gnupg/ directory from the top-level of this repo to the
        current working directory.

        :rtype: str
        :returns: The full path to the new gnupg home directory.
        """
        here         = os.getcwd()
        topDir       = here.rstrip('_trial_temp')
        gnupghome    = os.path.join(topDir, '.gnupg')
        gnupghomeNew = os.path.join(here, '.gnupg')

        if os.path.isdir(gnupghomeNew):
            shutil.rmtree(gnupghomeNew)

        shutil.copytree(gnupghome, gnupghomeNew)

        return gnupghomeNew

    def _writePassphraseToFile(self, passphrase, filename):
        """Write **passphrase** to the file at **filename**.

        :param str passphrase: The GnuPG passphase.
        :param str filename: The file to write the passphrase to.
        """
        fh = open(filename, 'w')
        fh.write(passphrase)
        fh.flush()
        fh.close()

    def setUp(self):
        """Create a config object and setup our gnupg home directory."""
        self.config = _createConfig()
        self.gnupghome = self._moveGnuPGHomedir()
        self.config.EMAIL_GPG_HOMEDIR = self.gnupghome

        self.passphraseFile = 'gpg-passphrase-file'
        self._writePassphraseToFile('sekrit', self.passphraseFile)

    def test_crypto_initializeGnuPG(self):
        """crypto.initializeGnuPG() should return a 2-tuple with a gpg object
        and a signing function.
        """
        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNotNone(gpg)
        self.assertIsNotNone(signfunc)

    def test_crypto_initializeGnuPG_disabled(self):
        """When EMAIL_GPG_SIGNING_ENABLED=False, crypto.initializeGnuPG()
        should return a 2-tuple of None.
        """
        self.config.EMAIL_GPG_SIGNING_ENABLED = False
        gpg, signfunc = crypto.initializeGnuPG(self.config)

        self.assertIsNone(gpg)
        self.assertIsNone(signfunc)

    def test_crypto_initializeGnuPG_no_secrets(self):
        """When the secring.gpg is missing, crypto.initializeGnuPG() should
        return a 2-tuple of None.
        """
        secring = os.path.join(self.gnupghome, 'secring.gpg')
        if os.path.isfile(secring):
            os.remove(secring)

        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNone(gpg)
        self.assertIsNone(signfunc)

    def test_crypto_initializeGnuPG_no_publics(self):
        """When the pubring.gpg is missing, crypto.initializeGnuPG() should
        return a 2-tuple of None.
        """
        pubring = os.path.join(self.gnupghome, 'pubring.gpg')
        if os.path.isfile(pubring):
            os.remove(pubring)

        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNone(gpg)
        self.assertIsNone(signfunc)

    def test_crypto_initializeGnuPG_with_passphrase(self):
        """crypto.initializeGnuPG() should initialize correctly when a
        passphrase is given but no passphrase is needed.
        """
        self.config.EMAIL_GPG_PASSPHRASE = 'password'
        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNotNone(gpg)
        self.assertIsNotNone(signfunc)

    def test_crypto_initializeGnuPG_with_passphrase_file(self):
        """crypto.initializeGnuPG() should initialize correctly when a
        passphrase file is given but no passphrase is needed.
        """
        self.config.EMAIL_GPG_PASSPHRASE_FILE = self.passphraseFile
        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNotNone(gpg)
        self.assertIsNotNone(signfunc)

    def test_crypto_initializeGnuPG_missing_passphrase_file(self):
        """crypto.initializeGnuPG() should initialize correctly if a passphrase
        file is given but that file is missing (when no passphrase is actually
        necessary).
        """
        self.config.EMAIL_GPG_PASSPHRASE_FILE = self.passphraseFile
        os.remove(self.passphraseFile)
        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNotNone(gpg)
        self.assertIsNotNone(signfunc)

    def test_crypto_initializeGnuPG_signingFunc(self):
        """crypto.initializeGnuPG() should return a signing function which
        produces OpenPGP signatures.
        """
        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNotNone(gpg)
        self.assertIsNotNone(signfunc)

        sig = signfunc("This is a test of the public broadcasting system.")
        print(sig)
        self.assertIsNotNone(sig)
        self.assertTrue(sig.startswith('-----BEGIN PGP SIGNED MESSAGE-----'))

    def test_crypto_initializeGnuPG_nonexistent_default_key(self):
        """When the key specified by EMAIL_GPG_PRIMARY_KEY_FINGERPRINT doesn't
        exist in the keyrings, crypto.initializeGnuPG() should return a 2-tuple
        of None.
        """
        self.config.EMAIL_GPG_PRIMARY_KEY_FINGERPRINT = 'A' * 40
        gpg, signfunc = crypto.initializeGnuPG(self.config)
        self.assertIsNone(gpg)
        self.assertIsNone(signfunc)


class RemovePKCS1PaddingTests(unittest.TestCase):
    """Unittests for :func:`bridgedb.crypto.removePKCS1Padding`."""

    def setUp(self):
        """This blob *is* actually a correctly formed PKCS#1 padded signature
        on the descriptor::

        @purpose bridge
        router ExhalesPeppier 118.16.116.176 35665 0 0
        or-address [eef2:d52a:cf1b:552f:375d:f8d0:a72b:e794]:35664
        platform Tor 0.2.4.5-alpha on Linux
        protocols Link 1 2 Circuit 1
        published 2014-11-03 21:21:43
        fingerprint FA04 5CFF AB95 BA20 C994 FE28 9B23 583E F80F 34DA
        uptime 10327748
        bandwidth 2247108152 2540209215 1954007088
        extra-info-digest 571BF23D8F24F052483C1333EBAE9B91E4A6F422
        onion-key
        -----BEGIN RSA PUBLIC KEY-----
        MIGJAoGBAK7+a033aUqc97SWFVGFwR3ybQ0jG1HTPtsv2/fUfZPwCaf21ly4zIvH
        9uNhtkcPH2p55X+n5M7OUaQawOzbwL4tSR9SLy9bGuZdWLbhu2GHQWmDkAB7BtHp
        UC+uGTN3jvQXEG2xlzpb+lOVUVNXLhL5kFmAXxL+iwN4TeEv/iCnAgMBAAE=
        -----END RSA PUBLIC KEY-----
        signing-key
        -----BEGIN RSA PUBLIC KEY-----
        MIGJAoGBANxmgJ6S3rBAGcvQu2tWBaHByJxeJkdGbxID2b8cITPaNmcl72e3Kd44
        GGIkoKhkX0SAO+i2U+Q41u/DPEBWLxhpl9GAFJZ10dcT18lL36yaK6FRDOcF9jx9
        0A023/kwXd7QQDWqP7Fso+141bzit6ENvNmE1mvEeIoAR+EpJB1tAgMBAAE=
        -----END RSA PUBLIC KEY-----
        contact Somebody <somebody@example.com>
        ntor-onion-key 0Mfi/Af7zLmdNdrmJyPbZxPJe7TZU/hV4Z865g3g+k4
        reject *:*
        router-signature
        -----BEGIN SIGNATURE-----
        PsGGIP+V9ZXWIHjK943CMAPem3kFbO9kt9rvrPhd64u0f7ytB/qZGaOg1IEWki1I
        f6ZNjrthxicm3vnEUdhpRsyn7MUFiQmqLjBfqdzh0GyfrtU5HHr7CBV3tuhgVhik
        uY1kPNo1C8wkmuy31H3V7NXj+etZuzZN66qL3BiQwa8=
        -----END SIGNATURE-----

        However, for the blob to be valid it would need to be converted from
        base64-decoded bytes to a long, then raised by the power of the public
        exponent within the ASN.1 DER decoded signing-key (mod that key's
        public modulus), then re-converted back into bytes before attempting
        to remove the PKCS#1 padding. (See
        :meth:`bridedb.bridges.Bridge._verifyExtraInfoSignature`.)
        """
        blob = ('PsGGIP+V9ZXWIHjK943CMAPem3kFbO9kt9rvrPhd64u0f7ytB/qZGaOg1IEWk'
                'i1If6ZNjrthxicm3vnEUdhpRsyn7MUFiQmqLjBfqdzh0GyfrtU5HHr7CBV3tu'
                'hgVhikuY1kPNo1C8wkmuy31H3V7NXj+etZuzZN66qL3BiQwa8=')
        self.blob = base64.b64decode(blob)

    def test_crypto_removePKCS1Padding_bad_padding(self):
        """removePKCS1Padding() with a blob with a bad PKCS#1 identifier mark
        should raise PKCS1PaddingError.
        """
        self.assertRaises(crypto.PKCS1PaddingError,
                          crypto.removePKCS1Padding,
                          self.blob)

    def test_crypto_removePKCS1Padding_missing_padding(self):
        """removePKCS1Padding() with a blob with a missing PKCS#1 identifier
        mark should raise PKCS1PaddingError.
        """
        self.assertRaises(crypto.PKCS1PaddingError,
                          crypto.removePKCS1Padding,
                          b'\x99' + self.blob)


class SSLVerifyingContextFactoryTests(unittest.TestCase,
                                      txtagent.FakeReactorAndConnectMixin):
    """Tests for :class:`bridgedb.crypto.SSLVerifyingContextFactory`."""

    _certificateText = (
        "-----BEGIN CERTIFICATE-----\n"
        "MIIEdjCCA16gAwIBAgIITcyHZlE/AhQwDQYJKoZIhvcNAQEFBQAwSTELMAkGA1UE\n"
        "BhMCVVMxEzARBgNVBAoTCkdvb2dsZSBJbmMxJTAjBgNVBAMTHEdvb2dsZSBJbnRl\n"
        "cm5ldCBBdXRob3JpdHkgRzIwHhcNMTQwMjEyMTUxMTE2WhcNMTQwNjEyMDAwMDAw\n"
        "WjBoMQswCQYDVQQGEwJVUzETMBEGA1UECAwKQ2FsaWZvcm5pYTEWMBQGA1UEBwwN\n"
        "TW91bnRhaW4gVmlldzETMBEGA1UECgwKR29vZ2xlIEluYzEXMBUGA1UEAwwOd3d3\n"
        "Lmdvb2dsZS5jb20wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCt3TOf\n"
        "VOf4vfy4IROcEyiFzAJA+B3xkMccwA4anaD6VyGSFglRn5Oht3t+G0Mnu/LMuGba\n"
        "EE6NEBEUEbH8KMlAcVRj58LoFIzulaRCdkVX7JK9R+kU05sggvIl1Q2quaWSjiMQ\n"
        "SpyvKz1I2cmU5Gm4MfW/66M5ZJO323VrV19ydrgAtdbNnvVj85asrSyzwEBNxzNC\n"
        "N6OQtOmTt4I7KLXqkROtTmTFvhAGBsvhG0hJZWhoP1aVsFO+KcE2OaIIxWQ4ckW7\n"
        "BJEgYaXfgHo01LdR55aevGUqLfsdyT+GMZrG9k7eqAw4cq3ML2Y6RiyzskqoQL30\n"
        "3OdYjKTIcU+i3BoFAgMBAAGjggFBMIIBPTAdBgNVHSUEFjAUBggrBgEFBQcDAQYI\n"
        "KwYBBQUHAwIwGQYDVR0RBBIwEIIOd3d3Lmdvb2dsZS5jb20waAYIKwYBBQUHAQEE\n"
        "XDBaMCsGCCsGAQUFBzAChh9odHRwOi8vcGtpLmdvb2dsZS5jb20vR0lBRzIuY3J0\n"
        "MCsGCCsGAQUFBzABhh9odHRwOi8vY2xpZW50czEuZ29vZ2xlLmNvbS9vY3NwMB0G\n"
        "A1UdDgQWBBQN7uQBzGDjvKRna111g9iPPtaXVTAMBgNVHRMBAf8EAjAAMB8GA1Ud\n"
        "IwQYMBaAFErdBhYbvPZotXb1gba7Yhq6WoEvMBcGA1UdIAQQMA4wDAYKKwYBBAHW\n"
        "eQIFATAwBgNVHR8EKTAnMCWgI6Ahhh9odHRwOi8vcGtpLmdvb2dsZS5jb20vR0lB\n"
        "RzIuY3JsMA0GCSqGSIb3DQEBBQUAA4IBAQBrVp/xys2ABQvWPxpVrYaXiaoBXdxu\n"
        "RVVXp5Lyu8IipKqFJli81hOX9eqPG7biYeph9HiKnW31xsXebaVlWWL3NXOh5X83\n"
        "wpzozL0AkxskTMHQknrbIGLtmG67H71aKYyCthHEjawLmYjjvkcF6f9fKdYENM4C\n"
        "skz/yjtlPBQFAuT6J9w0b3qtc42sHNlpgIOdIRQc2YCD0p6jAo+wKjoRuRu3ILKj\n"
        "oCVrOPbDMPN4a2gSmK8Ur0aHuEpcNghg6HJsVSANokIIwQ/r4niqL5yotsangP/5\n"
        "rR97EIYKFz7C6LMy/PIe8xFTIyKMtM59IcpUDIwCLlM9JtNdwN4VpyKy\n"
        "-----END CERTIFICATE-----\n")

    def setUp(self):
        """Create a fake reactor for these tests."""
        self.reactor = self.Reactor()
        self.url = 'https://www.example.com/someresource.html#andatag'

    def test_getHostnameFromURL(self):
        """``getHostnameFromURL()`` should return a hostname from a URI."""
        if _twistedversion >= Version('twisted', 14, 0, 0):
            raise unittest.SkipTest(
                ("The SSLVerifyingContextFactory is no longer necessary in "
                 "Twisted>=14.0.0, because the way in which TLS certificates "
                 "are checked now includes certificate pinning, and the "
                 "SSLVerifyingContextFactory only implemented strict hostname "
                 "checking."))

        agent = txrecaptcha._getAgent(self.reactor, self.url)
        contextFactory = agent._contextFactory
        self.assertRegexpMatches(contextFactory.hostname,
                                 '.*www\.example\.com')

    def test_verifyHostname_mismatching(self):
        """Check that ``verifyHostname()`` returns ``False`` when the
        ``SSLVerifyingContextFactory.hostname`` does not match the one found
        in the level 0 certificate subject CN.
        """
        if _twistedversion >= Version('twisted', 14, 0, 0):
            raise unittest.SkipTest(
                ("The SSLVerifyingContextFactory is no longer necessary in "
                 "Twisted>=14.0.0, because the way in which TLS certificates "
                 "are checked now includes certificate pinning, and the "
                 "SSLVerifyingContextFactory only implemented strict hostname "
                 "checking."))

        agent = txrecaptcha._getAgent(self.reactor, self.url)
        contextFactory = agent._contextFactory
        x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM,
                                               self._certificateText)
        conn = DummyEndpoint()
        result = contextFactory.verifyHostname(conn, x509, 0, 0, True)
        self.assertIs(result, False)

    def test_verifyHostname_matching(self):
        """Check that ``verifyHostname()`` returns ``True`` when the
        ``SSLVerifyingContextFactory.hostname`` matches the one found in the
        level 0 certificate subject CN.
        """
        hostname = 'www.google.com'
        url = 'https://' + hostname + '/recaptcha'
        contextFactory = crypto.SSLVerifyingContextFactory(url)
        self.assertEqual(contextFactory.hostname, hostname)

        x509 = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM,
                                               self._certificateText)
        conn = DummyEndpoint()
        result = contextFactory.verifyHostname(conn, x509, 0, 0, True)
        self.assertTrue(result)

    def test_getContext(self):
        """The context factory's ``getContext()`` method should produce an
        ``OpenSSL.SSL.Context`` object.
        """
        contextFactory = crypto.SSLVerifyingContextFactory(self.url)
        self.assertIsInstance(contextFactory.getContext(),
                              OpenSSL.SSL.Context)



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           please also see AUTHORS file
# :copyright: (c) 2013-2015 Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
#             (c) 2007-2015, all entities within the AUTHORS file
# :license: 3-clause BSD, see included LICENSE for information

"""Tests for :mod:`bridgedb.https.distributor`."""

from __future__ import print_function

import ipaddr
import logging
import random

from twisted.trial import unittest

from bridgedb.Bridges import BridgeRing
from bridgedb.Bridges import BridgeRingParameters
from bridgedb.filters import byIPv4
from bridgedb.filters import byIPv6
from bridgedb.https import distributor
from bridgedb.https.request import HTTPSBridgeRequest
from bridgedb.proxy import ProxySet

from .util import randomValidIPv4String
from .util import generateFakeBridges
from .https_helpers import DummyRequest

logging.disable(50)


BRIDGES = generateFakeBridges()


class HTTPSDistributorTests(unittest.TestCase):
    """Tests for :class:`HTTPSDistributor`."""

    def setUp(self):
        self.key = 'aQpeOFIj8q20s98awfoiq23rpOIjFaqpEWFoij1X'
        self.bridges = BRIDGES

    def tearDown(self):
        """Reset all bridge blocks in between test method runs."""
        for bridge in self.bridges:
            bridge._blockedIn = {}

    def coinFlip(self):
        return bool(random.getrandbits(1))

    def randomClientRequest(self):
        bridgeRequest = HTTPSBridgeRequest(addClientCountryCode=False)
        bridgeRequest.client = randomValidIPv4String()
        bridgeRequest.isValid(True)
        bridgeRequest.generateFilters()
        return bridgeRequest

    def randomClientRequestForNotBlockedIn(self, cc):
        httpRequest = DummyRequest([''])
        httpRequest.args.update({'unblocked': [cc]})
        bridgeRequest = self.randomClientRequest()
        bridgeRequest.withoutBlockInCountry(httpRequest)
        bridgeRequest.generateFilters()
        return bridgeRequest

    def test_HTTPSDistributor_init_with_proxies(self):
        """The HTTPSDistributor, when initialised with proxies, should add an
        extra hashring for proxy users.
        """
        dist = distributor.HTTPSDistributor(3, self.key, ProxySet(['1.1.1.1', '2.2.2.2']))
        self.assertIsNotNone(dist.proxies)
        self.assertGreater(dist.proxySubring, 0)
        self.assertEqual(dist.proxySubring, 4)
        self.assertEqual(dist.totalSubrings, 4)

    def test_HTTPSDistributor_bridgesPerResponse_120(self):
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:120]]
        self.assertEqual(dist.bridgesPerResponse(), 3)

    def test_HTTPSDistributor_bridgesPerResponse_100(self):
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.hashring.insert(bridge) for bridge in self.bridges[:100]]
        self.assertEqual(dist.bridgesPerResponse(), 3)

    def test_HTTPSDistributor_bridgesPerResponse_50(self):
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:60]]
        self.assertEqual(dist.bridgesPerResponse(), 2)

    def test_HTTPSDistributor_bridgesPerResponse_15(self):
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:15]]
        self.assertEqual(dist.bridgesPerResponse(), 1)

    def test_HTTPSDistributor_bridgesPerResponse_100_max_5(self):
        dist = distributor.HTTPSDistributor(3, self.key)
        dist._bridgesPerResponseMax = 5
        [dist.insert(bridge) for bridge in self.bridges[:100]]
        self.assertEqual(dist.bridgesPerResponse(), 5)

    def test_HTTPSDistributor_getSubnet_usingProxy(self):
        """HTTPSDistributor.getSubnet(usingProxy=True) should return a proxy
        group number.
        """
        clientRequest = self.randomClientRequest()
        expectedGroup = (int(ipaddr.IPAddress(clientRequest.client)) % 4) + 1
        subnet = distributor.HTTPSDistributor.getSubnet(clientRequest.client, usingProxy=True)
        self.assertTrue(subnet.startswith('proxy-group-'))
        self.assertEqual(int(subnet[-1]), expectedGroup)

    def test_HTTPSDistributor_mapSubnetToSubring_usingProxy(self):
        """HTTPSDistributor.mapSubnetToSubring() when the client was using a
        proxy should map the client to the proxy subhashring.
        """
        dist = distributor.HTTPSDistributor(3, self.key, ProxySet(['1.1.1.1', '2.2.2.2']))
        subnet = 'proxy-group-3'
        subring = dist.mapSubnetToSubring(subnet, usingProxy=True)
        self.assertEqual(subring, dist.proxySubring)

    def test_HTTPSDistributor_mapSubnetToSubring_with_proxies(self):
        """HTTPSDistributor.mapSubnetToSubring() when the client wasn't using
        a proxy, but the distributor does have some known proxies and a
        proxySubring, should not map the client to the proxy subhashring.
        """
        dist = distributor.HTTPSDistributor(3, self.key, ProxySet(['1.1.1.1', '2.2.2.2']))
        # Note that if they were actually from a proxy, their subnet would be
        # something like "proxy-group-3".
        subnet = '15.1.0.0/16'
        subring = dist.mapSubnetToSubring(subnet, usingProxy=False)
        self.assertNotEqual(subring, dist.proxySubring)

    def test_HTTPSDistributor_prepopulateRings_with_proxies(self):
        """An HTTPSDistributor with proxies should prepopulate two extra
        subhashrings (one for each of HTTP-Proxy-IPv4 and HTTP-Proxy-IPv6).
        """
        dist = distributor.HTTPSDistributor(3, self.key, ProxySet(['1.1.1.1', '2.2.2.2']))
        [dist.insert(bridge) for bridge in self.bridges]
        dist.prepopulateRings()
        self.assertEqual(len(dist.hashring.filterRings), 8)

    def test_HTTPSDistributor_prepopulateRings_without_proxies(self):
        """An HTTPSDistributor without proxies should prepopulate
        totalSubrings * 2 subrings.
        """
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.insert(bridge) for bridge in self.bridges]
        dist.prepopulateRings()
        self.assertEqual(len(dist.hashring.filterRings), 6)

        ipv4subrings = []
        ipv6subrings = []

        for subringName, (filters, subring) in dist.hashring.filterRings.items():
            if 'IPv4' in subringName:
                ipv6subrings.append(subring)
            if 'IPv6' in subringName:
                ipv6subrings.append(subring)

        self.assertEqual(len(ipv4subrings), len(ipv6subrings))

    def test_HTTPSDistributor_getBridges_with_blocked_bridges(self):
        dist = distributor.HTTPSDistributor(1, self.key)
        bridges = self.bridges[:]

        for bridge in bridges:
            bridge.setBlockedIn('cn')

        [dist.insert(bridge) for bridge in bridges]

        for _ in range(5):
            clientRequest1 = self.randomClientRequestForNotBlockedIn('cn')
            b = dist.getBridges(clientRequest1, 1)
            self.assertEqual(len(b), 0)

            clientRequest2 = self.randomClientRequestForNotBlockedIn('ir')
            b = dist.getBridges(clientRequest2, 1)
            self.assertEqual(len(b), 3)

    def test_HTTPSDistributor_getBridges_with_some_blocked_bridges(self):
        dist = distributor.HTTPSDistributor(1, self.key)
        bridges = self.bridges[:]

        blockedCN = []
        blockedIR = []

        for bridge in bridges:
            if self.coinFlip():
                bridge.setBlockedIn('cn')
                blockedCN.append(bridge.fingerprint)

            if self.coinFlip():
                bridge.setBlockedIn('ir')
                blockedIR.append(bridge.fingerprint)

        [dist.insert(bridge) for bridge in bridges]

        for _ in range(5):
            clientRequest1 = self.randomClientRequestForNotBlockedIn('cn')
            bridges = dist.getBridges(clientRequest1, 1)
            for b in bridges:
                self.assertFalse(b.isBlockedIn('cn'))
                self.assertNotIn(b.fingerprint, blockedCN)
            # The client *should* have gotten some bridges still.
            self.assertGreater(len(bridges), 0)

            clientRequest2 = self.randomClientRequestForNotBlockedIn('ir')
            bridges = dist.getBridges(clientRequest2, 1)
            for b in bridges:
                self.assertFalse(b.isBlockedIn('ir'))
                self.assertNotIn(b.fingerprint, blockedIR)
            self.assertGreater(len(bridges), 0)

    def test_HTTPSDistributor_getBridges_with_varied_blocked_bridges(self):
        dist = distributor.HTTPSDistributor(1, self.key)
        bridges = self.bridges[:]

        for bridge in bridges:
            # Pretend that China blocks all vanilla bridges:
            bridge.setBlockedIn('cn', methodname='vanilla')
            # Pretend that China blocks all obfs2:
            bridge.setBlockedIn('cn', methodname='obfs2')
            # Pretend that China blocks some obfs3:
            if self.coinFlip():
                bridge.setBlockedIn('cn', methodname='obfs3')

        [dist.insert(bridge) for bridge in bridges]

        for i in xrange(5):
            bridgeRequest1 = self.randomClientRequestForNotBlockedIn('cn')
            bridgeRequest1.transports.append('obfs2')
            bridgeRequest1.generateFilters()
            # We shouldn't get any obfs2 bridges, since they're all blocked in
            # China:
            bridges = dist.getBridges(bridgeRequest1, "faketimestamp")
            self.assertEqual(len(bridges), 0)

            bridgeRequest2 = self.randomClientRequestForNotBlockedIn('cn')
            bridgeRequest2.transports.append('obfs3')
            bridgeRequest2.generateFilters()
            # We probably will get at least one bridge back!  It's pretty
            # unlikely to lose a coin flip 500 times in a row.
            bridges = dist.getBridges(bridgeRequest2, "faketimestamp")
            self.assertGreater(len(bridges), 0)

            bridgeRequest3 = self.randomClientRequestForNotBlockedIn('nl')
            bridgeRequest3.transports.append('obfs3')
            bridgeRequest3.generateFilters()
            # We should get bridges, since obfs3 isn't blocked in netherlands:
            bridges = dist.getBridges(bridgeRequest3, "faketimestamp")
            self.assertGreater(len(bridges), 0)

    def test_HTTPSDistributor_getBridges_with_proxy_and_nonproxy_users(self):
        """An HTTPSDistributor should give separate bridges to proxy users."""
        proxies = ProxySet(['.'.join(['1.1.1', str(x)]) for x in range(1, 256)])
        dist = distributor.HTTPSDistributor(3, self.key, proxies)
        [dist.insert(bridge) for bridge in self.bridges]

        for _ in range(10):
            bridgeRequest1 = self.randomClientRequest()
            bridgeRequest1.client = '.'.join(['1.1.1', str(random.randrange(1, 255))])

            bridgeRequest2 = self.randomClientRequest()
            bridgeRequest2.client = '.'.join(['9.9.9', str(random.randrange(1, 255))])

            n1 = dist.getBridges(bridgeRequest1, 1)
            n2 = dist.getBridges(bridgeRequest2, 1)

            self.assertGreater(len(n1), 0)
            self.assertGreater(len(n2), 0)

            for b in n1:
                self.assertNotIn(b, n2)
            for b in n2:
                self.assertNotIn(b, n1)

    def test_HTTPSDistributor_getBridges_same_bridges_to_same_client(self):
        """The same client asking for bridges from the HTTPSDistributor
        multiple times in a row should get the same bridges in response each
        time.
        """
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:250]]

        bridgeRequest = self.randomClientRequest()
        responses = {}
        for i in range(5):
            responses[i] = dist.getBridges(bridgeRequest, 1)
        for i in range(4):
            self.assertItemsEqual(responses[i], responses[i+1])

    def test_HTTPSDistributor_getBridges_with_BridgeRingParameters(self):
       param = BridgeRingParameters(needPorts=[(443, 1)])
       dist = distributor.HTTPSDistributor(3, self.key, answerParameters=param)

       bridges = self.bridges[:32]
       for b in self.bridges:
           b.orPort = 443

       [dist.insert(bridge) for bridge in bridges]
       [dist.insert(bridge) for bridge in self.bridges[:250]]

       for _ in xrange(32):
           bridgeRequest = self.randomClientRequest()
           answer = dist.getBridges(bridgeRequest, 1)
           count = 0
           fingerprints = {}
           for bridge in answer:
               fingerprints[bridge.identity] = 1
               if bridge.orPort == 443:
                   count += 1
           self.assertEquals(len(fingerprints), len(answer))
           self.assertGreater(len(fingerprints), 0)
           self.assertTrue(count >= 1)

    def test_HTTPSDistributor_getBridges_ipv4_ipv6(self):
        """Asking for bridge addresses which are simultaneously IPv4 and IPv6
        (in that order) should return IPv4 bridges.
        """
        dist = distributor.HTTPSDistributor(1, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:250]]

        bridgeRequest = self.randomClientRequest()
        bridgeRequest.withIPv4()
        bridgeRequest.filters.append(byIPv6)
        bridgeRequest.generateFilters()

        bridges = dist.getBridges(bridgeRequest, 1)
        self.assertEqual(len(bridges), 3)

        bridge = random.choice(bridges)
        bridgeLine = bridge.getBridgeLine(bridgeRequest)
        addrport, fingerprint = bridgeLine.split()
        address, port = addrport.rsplit(':', 1)
        address = address.strip('[]')
        self.assertIsInstance(ipaddr.IPAddress(address), ipaddr.IPv4Address)
        self.assertIsNotNone(byIPv4(random.choice(bridges)))

    def test_HTTPSDistributor_getBridges_ipv6_ipv4(self):
        """Asking for bridge addresses which are simultaneously IPv6 and IPv4
        (in that order) should return IPv6 bridges.
        """
        dist = distributor.HTTPSDistributor(1, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:250]]

        bridgeRequest = self.randomClientRequest()
        bridgeRequest.withIPv6()
        bridgeRequest.generateFilters()
        bridgeRequest.filters.append(byIPv4)

        bridges = dist.getBridges(bridgeRequest, 1)
        self.assertEqual(len(bridges), 3)

        bridge = random.choice(bridges)
        bridgeLine = bridge.getBridgeLine(bridgeRequest)
        addrport, fingerprint = bridgeLine.split()
        address, port = addrport.rsplit(':', 1)
        address = address.strip('[]')
        self.assertIsInstance(ipaddr.IPAddress(address), ipaddr.IPv6Address)
        self.assertIsNotNone(byIPv6(random.choice(bridges)))

    def test_HTTPSDistributor_getBridges_ipv6(self):
        """A request for IPv6 bridges should return IPv6 bridges."""
        dist = distributor.HTTPSDistributor(3, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:250]]

        for i in xrange(500):
            bridgeRequest = self.randomClientRequest()
            bridgeRequest.withIPv6()
            bridgeRequest.generateFilters()

            bridges = dist.getBridges(bridgeRequest, "faketimestamp")
            self.assertTrue(type(bridges) is list)
            self.assertGreater(len(bridges), 0)

            bridge = random.choice(bridges)
            bridgeLine = bridge.getBridgeLine(bridgeRequest)
            addrport, fingerprint = bridgeLine.split()
            address, port = addrport.rsplit(':', 1)
            address = address.strip('[]')
            self.assertIsInstance(ipaddr.IPAddress(address), ipaddr.IPv6Address)
            self.assertIsNotNone(byIPv6(random.choice(bridges)))

    def test_HTTPSDistributor_getBridges_ipv4(self):
        """A request for IPv4 bridges should return IPv4 bridges."""
        dist = distributor.HTTPSDistributor(1, self.key)
        [dist.insert(bridge) for bridge in self.bridges[:250]]

        for i in xrange(500):
            bridgeRequest = self.randomClientRequest()
            bridgeRequest.generateFilters()

            bridges = dist.getBridges(bridgeRequest, "faketimestamp")
            self.assertTrue(type(bridges) is list)
            self.assertGreater(len(bridges), 0)

            bridge = random.choice(bridges)
            bridgeLine = bridge.getBridgeLine(bridgeRequest)
            addrport, fingerprint = bridgeLine.split()
            address, port = addrport.rsplit(':', 1)
            self.assertIsInstance(ipaddr.IPAddress(address), ipaddr.IPv4Address)
            self.assertIsNotNone(byIPv4(random.choice(bridges)))



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_https_distributor -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Nick Mathewson
#           Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
#           Matthew Finkel 0x017DD169EA793BE2 <sysrqb@torproject.org>
# :copyright: (c) 2013-2015, Isis Lovecruft
#             (c) 2013-2015, Matthew Finkel
#             (c) 2007-2015, The Tor Project, Inc.
# :license: see LICENSE for licensing information

"""
bridgedb.https.distributor
==========================

A Distributor that hands out bridges through a web interface.

.. inheritance-diagram:: HTTPSDistributor
    :parts: 1
"""

import ipaddr
import logging

import bridgedb.Storage

from bridgedb import proxy
from bridgedb.Bridges import BridgeRing
from bridgedb.Bridges import FilteredBridgeSplitter
from bridgedb.crypto import getHMAC
from bridgedb.crypto import getHMACFunc
from bridgedb.distribute import Distributor
from bridgedb.filters import byIPv4
from bridgedb.filters import byIPv6
from bridgedb.filters import byFilters
from bridgedb.filters import bySubring


class HTTPSDistributor(Distributor):
    """A Distributor that hands out bridges based on the IP address of an
    incoming request and the current time period.

    :type proxies: :class:`~bridgedb.proxies.ProxySet`
    :ivar proxies: All known proxies, which we treat differently. See
        :param:`proxies`.
    :type hashring: :class:`bridgedb.Bridges.FilteredBridgeSplitter`
    :ivar hashring: A hashring that assigns bridges to subrings with fixed
        proportions. Used to assign bridges into the subrings of this
        distributor.
    """

    def __init__(self, totalSubrings, key, proxies=None, answerParameters=None):
        """Create a Distributor that decides which bridges to distribute based
        upon the client's IP address and the current time.

        :param int totalSubrings: The number of subhashrings to group clients
            into. Note that if ``PROXY_LIST_FILES`` is set in bridgedb.conf,
            then the actual number of clusters is one higher than
            ``totalSubrings``, because the set of all known open proxies is
            given its own subhashring.
        :param bytes key: The master HMAC key for this distributor. All added
            bridges are HMACed with this key in order to place them into the
            hashrings.
        :type proxies: :class:`~bridgedb.proxy.ProxySet`
        :param proxies: A :class:`bridgedb.proxy.ProxySet` containing known
            Tor Exit relays and other known proxies.  These will constitute
            the extra cluster, and any client requesting bridges from one of
            these **proxies** will be distributed bridges from a separate
            subhashring that is specific to Tor/proxy users.
        :type answerParameters: :class:`bridgedb.Bridges.BridgeRingParameters`
        :param answerParameters: A mechanism for ensuring that the set of
            bridges that this distributor answers a client with fit certain
            parameters, i.e. that an answer has "at least two obfsproxy
            bridges" or "at least one bridge on port 443", etc.
        """
        super(HTTPSDistributor, self).__init__(key)
        self.totalSubrings = totalSubrings
        self.answerParameters = answerParameters

        if proxies:
            logging.info("Added known proxies to HTTPS distributor...")
            self.proxies = proxies
            self.totalSubrings += 1
            self.proxySubring = self.totalSubrings
        else:
            logging.warn("No known proxies were added to HTTPS distributor!")
            self.proxies = proxy.ProxySet()
            self.proxySubring = 0

        self.ringCacheSize = self.totalSubrings * 3

        key2 = getHMAC(key, "Assign-Bridges-To-Rings")
        key3 = getHMAC(key, "Order-Areas-In-Rings")
        key4 = getHMAC(key, "Assign-Areas-To-Rings")

        self._clientToPositionHMAC = getHMACFunc(key3, hex=False)
        self._subnetToSubringHMAC = getHMACFunc(key4, hex=True)
        self.hashring = FilteredBridgeSplitter(key2, self.ringCacheSize)
        self.name = 'HTTPS'
        logging.debug("Added %s to %s distributor." %
                      (self.hashring.__class__.__name__, self.name))

    def bridgesPerResponse(self, hashring=None):
        return super(HTTPSDistributor, self).bridgesPerResponse(hashring)

    @classmethod
    def getSubnet(cls, ip, usingProxy=False, proxySubnets=4):
        """Map all clients whose **ip**s are within the same subnet to the same
        arbitrary string.

        .. hint:: For non-proxy IP addresses, any two IPv4 addresses within
            the same ``/16`` subnet, or any two IPv6 addresses in the same
            ``/32`` subnet, will get the same string.

        Subnets for this distributor are grouped into the number of rings
        specified by the ``N_IP_CLUSTERS`` configuration option, such that
        Alice (with the address ``1.2.3.4`` and Bob (with the address
        ``1.2.178.234``) are placed within the same cluster, but Carol (with
        address ``1.3.11.33``) *might* end up in a different cluster.

        >>> from bridgedb.https.distributor import HTTPSDistributor
        >>> HTTPSDistributor.getSubnet('1.2.3.4')
        '1.2.0.0/16'
        >>> HTTPSDistributor.getSubnet('1.2.211.154')
        '1.2.0.0/16'
        >>> HTTPSDistributor.getSubnet('2001:f::bc1:b13:2808')
        '2001:f::/32'
        >>> HTTPSDistributor.getSubnet('2a00:c98:2030:a020:2::42')
        '2a00:c98::/32'

        :param str ip: A string representing an IPv4 or IPv6 address.
        :param bool usingProxy: Set to ``True`` if the client was using one of
            the known :data:`proxies`.
        :param int proxySubnets: Place Tor/proxy users into this number of
            "subnet" groups.  This means that no matter how many different Tor
            Exits or proxies a client uses, the most they can ever get is
            **proxySubnets** different sets of bridge lines (per interval).
            This parameter only has any effect when **usingProxy** is ``True``.
        :rtype: str
        :returns: The appropriately sized CIDR subnet representation of the **ip**.
        """
        if not usingProxy:
            # We aren't using bridgedb.parse.addr.isIPAddress(ip,
            # compressed=False) here because adding the string "False" into
            # the map would land any and all clients whose IP address appeared
            # to be invalid at the same position in a hashring.
            address = ipaddr.IPAddress(ip)
            if address.version == 6:
                truncated = ':'.join(address.exploded.split(':')[:2])
                subnet = str(ipaddr.IPv6Network(truncated + "::/32"))
            else:
                truncated = '.'.join(address.exploded.split('.')[:2])
                subnet = str(ipaddr.IPv4Network(truncated + '.0.0/16'))
        else:
            group = (int(ipaddr.IPAddress(ip)) % 4) + 1
            subnet = "proxy-group-%d" % group

        logging.debug("Client IP was within area: %s" % subnet)
        return subnet

    def mapSubnetToSubring(self, subnet, usingProxy=False):
        """Determine the correct subhashring for a client, based upon the
        **subnet**.

        :param str subnet: The subnet which contains the client's IP.  See
            :staticmethod:`getSubnet`.
        :param bool usingProxy: Set to ``True`` if the client was using one of
            the known :data:`proxies`.
        """
        # If the client wasn't using a proxy, select the client's subring
        # based upon the client's subnet (modulo the total subrings):
        if not usingProxy:
            mod = self.totalSubrings
            # If there is a proxy subring, don't count it for the modulus:
            if self.proxySubring:
                mod -= 1
            return (int(self._subnetToSubringHMAC(subnet)[:8], 16) % mod) + 1
        else:
            return self.proxySubring

    def mapClientToHashringPosition(self, interval, subnet):
        """Map the client to a position on a (sub)hashring, based upon the
        **interval** which the client's request occurred within, as well as
        the **subnet** of the client's IP address.

        .. note:: For an explanation of how **subnet** is determined, see
            :staticmethod:`getSubnet`.

        :param str interval: The interval which this client's request for
            bridges took place within.
        :param str subnet: A string representing the subnet containing the
            client's IP address.
        :rtype: int
        :returns: The results of keyed HMAC, which should determine the
            client's position in a (sub)hashring of bridges (and thus
            determine which bridges they receive).
        """
        position = "<%s>%s" % (interval, subnet)
        mapping = self._clientToPositionHMAC(position)
        return mapping

    def prepopulateRings(self):
        """Prepopulate this distributor's hashrings and subhashrings with
        bridges.

        The hashring structure for this distributor is influenced by the
        ``N_IP_CLUSTERS`` configuration option, as well as the number of
        ``PROXY_LIST_FILES``.

        Essentially, :data:`totalSubrings` is set to the specified
        ``N_IP_CLUSTERS``.  All of the ``PROXY_LIST_FILES``, plus the list of
        Tor Exit relays (downloaded into memory with :script:`get-tor-exits`),
        are stored in :data:`proxies`, and the latter is added as an
        additional cluster (such that :data:`totalSubrings` becomes
        ``N_IP_CLUSTERS + 1``).  The number of subhashrings which this
        :class:`Distributor` has active in its hashring is then
        :data:`totalSubrings`, where the last cluster is reserved for all
        :data:`proxies`.

        As an example, if BridgeDB was configured with ``N_IP_CLUSTERS=4`` and
        ``PROXY_LIST_FILES=["open-socks-proxies.txt"]``, then the total number
        of subhashrings is five — four for the "clusters", and one for the
        :data:`proxies`. Thus, the resulting hashring-subhashring structure
        would look like:

        +------------------+---------------------------------------------------+-------------+
        |                  |               Directly connecting users           | Tor / known |
        |                  |                                                   | proxy users |
        +------------------+------------+------------+------------+------------+-------------+
        | Clusters         | Cluster-1  | Cluster-2  | Cluster-3  | Cluster-4  | Cluster-5   |
        +==================+============+============+============+============+=============+
        | Subhashrings     |            |            |            |            |             |
        | (total, assigned)| (5,1)      | (5,2)      | (5,3)      | (5,4)      | (5,5)       |
        +------------------+------------+------------+------------+------------+-------------+
        | Filtered         | (5,1)-IPv4 | (5,2)-IPv4 | (5,3)-IPv4 | (5,4)-IPv4 | (5,5)-IPv4  |
        | Subhashrings     |            |            |            |            |             |
        | bBy requested    +------------+------------+------------+------------+-------------+
        | bridge type)     | (5,1)-IPv6 | (5,2)-IPv6 | (5,3)-IPv6 | (5,4)-IPv6 | (5,5)-IPv6  |
        |                  |            |            |            |            |             |
        +------------------+------------+------------+------------+------------+-------------+

        The "filtered subhashrings" are essentially filtered copies of their
        respective subhashring, such that they only contain bridges which
        support IPv4 or IPv6, respectively.  Additionally, the contents of
        ``(5,1)-IPv4`` and ``(5,1)-IPv6`` sets are *not* disjoint.

        Thus, in this example, we end up with **10 total subhashrings**.
        """
        logging.info("Prepopulating %s distributor hashrings..." % self.name)

        for filterFn in [byIPv4, byIPv6]:
            for subring in range(1, self.totalSubrings + 1):
                filters = self._buildHashringFilters([filterFn,], subring)
                key1 = getHMAC(self.key, "Order-Bridges-In-Ring-%d" % subring)
                ring = BridgeRing(key1, self.answerParameters)
                # For consistency with previous implementation of this method,
                # only set the "name" for "clusters" which are for this
                # distributor's proxies:
                if subring == self.proxySubring:
                    ring.setName('{0} Proxy Ring'.format(self.name))
                self.hashring.addRing(ring, filters, byFilters(filters),
                                      populate_from=self.hashring.bridges)

    def insert(self, bridge):
        """Assign a bridge to this distributor."""
        self.hashring.insert(bridge)

    def _buildHashringFilters(self, previousFilters, subring):
        f = bySubring(self.hashring.hmac, subring, self.totalSubrings)
        previousFilters.append(f)
        return frozenset(previousFilters)

    def getBridges(self, bridgeRequest, interval):
        """Return a list of bridges to give to a user.

        :type bridgeRequest: :class:`bridgedb.https.request.HTTPSBridgeRequest`
        :param bridgeRequest: A :class:`~bridgedb.bridgerequest.BridgeRequestBase`
            with the :data:`~bridgedb.bridgerequest.BridgeRequestBase.client`
            attribute set to a string containing the client's IP address.
        :param str interval: The time period when we got this request.  This
            can be any string, so long as it changes with every period.
        :rtype: list
        :return: A list of :class:`~bridgedb.Bridges.Bridge`s to include in
            the response. See
            :meth:`bridgedb.https.server.WebResourceBridges.getBridgeRequestAnswer`
            for an example of how this is used.
        """
        logging.info("Attempting to get bridges for %s..." % bridgeRequest.client)

        if not len(self.hashring):
            logging.warn("Bailing! Hashring has zero bridges!")
            return []

        usingProxy = False

        # First, check if the client's IP is one of the known :data:`proxies`:
        if bridgeRequest.client in self.proxies:
            # The tag is a tag applied to a proxy IP address when it is added
            # to the bridgedb.proxy.ProxySet. For Tor Exit relays, the default
            # is 'exit_relay'. For other proxies loaded from the
            # PROXY_LIST_FILES config option, the default tag is the full
            # filename that the IP address originally came from.
            usingProxy = True
            tag = self.proxies.getTag(bridgeRequest.client)
            logging.info("Client was from known proxy (tag: %s): %s" %
                         (tag, bridgeRequest.client))

        subnet = self.getSubnet(bridgeRequest.client, usingProxy)
        subring = self.mapSubnetToSubring(subnet, usingProxy)
        position = self.mapClientToHashringPosition(interval, subnet)
        filters = self._buildHashringFilters(bridgeRequest.filters, subring)

        logging.debug("Client request within time interval: %s" % interval)
        logging.debug("Assigned client to subhashring %d/%d" % (subring, self.totalSubrings))
        logging.debug("Assigned client to subhashring position: %s" % position.encode('hex'))
        logging.debug("Total bridges: %d" % len(self.hashring))
        logging.debug("Bridge filters: %s" % ' '.join([x.func_name for x in filters]))

        # Check wheth we have a cached copy of the hashring:
        if filters in self.hashring.filterRings.keys():
            logging.debug("Cache hit %s" % filters)
            _, ring = self.hashring.filterRings[filters]
        # Otherwise, construct a new hashring and populate it:
        else:
            logging.debug("Cache miss %s" % filters)
            key1 = getHMAC(self.key, "Order-Bridges-In-Ring-%d" % subring)
            ring = BridgeRing(key1, self.answerParameters)
            self.hashring.addRing(ring, filters, byFilters(filters),
                                  populate_from=self.hashring.bridges)

        # Determine the appropriate number of bridges to give to the client:
        returnNum = self.bridgesPerResponse(ring)
        answer = ring.getBridges(position, returnNum)

        return answer



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2013-2015, Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
# :license: 3-Clause BSD, see LICENSE for licensing information

"""Unittests for the bridgedb.txrecaptcha module."""

import logging
import os

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.base import DelayedCall
from twisted.internet.error import ConnectionDone
from twisted.internet.error import ConnectionLost
from twisted.internet.error import ConnectionRefusedError
from twisted.test import proto_helpers
from twisted.trial import unittest
from twisted.trial.unittest import SkipTest
from twisted.python import failure
from twisted.web.client import ResponseDone
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer

from zope.interface.verify import verifyObject

from bridgedb import txrecaptcha


logging.disable(50)

# Set ``DelayedCall.debug=True``, because the following traceback was occuring:
#
# Traceback (most recent call last):
# Failure: twisted.trial.util.DirtyReactorAggregateError: Reactor was unclean.
# DelayedCalls: (set twisted.internet.base.DelayedCall.debug = True to debug)
# <DelayedCall 0x1ba5b90 [29.991571188s] called=0 cancelled=0
#     Client.failIfNotConnected(TimeoutError('',))>
# <DelayedCall 0x1baa3f8 [59.9993360043s] called=0 cancelled=0
#     ThreadedResolver._cleanup('www.google.com', <Deferred at 0x1baa320>)>
DelayedCall.debug = True


class MockResponse(object):
    """Fake :api:`twisted.internet.interfaces.IResponse` for testing readBody
    that just captures the protocol passed to deliverBody.

    :ivar protocol: After :meth:`deliverBody` is called, the protocol it was
        called with.
    """
    code = 200
    phrase = "OK"

    def __init__(self, headers=None):
        """Create a mock response.

        :type headers: :api:`twisted.web.http_headers.Headers`
        :param headers: The headers for this response.  If ``None``, an empty
            ``Headers`` instance will be used.
        """
        if headers is None:
            headers = Headers()
        self.headers = headers

    def deliverBody(self, protocol):
        """Just record the given protocol without actually delivering anything
        to it.
        """
        self.protocol = protocol


class RecaptchaResponseProtocolTests(unittest.TestCase):
    """Tests for bridgedb.txrecaptcha.RecaptchaResponseProtocol."""

    def setUp(self):
        """Setup the tests."""
        self.finished = defer.Deferred()
        self.proto = txrecaptcha.RecaptchaResponseProtocol(self.finished)

    def _test(self, responseBody, connCloseError):
        """Deliver the **responseBody** to
        ``RecaptchaResponseProtocol.dataReceived``, and then lose the transport
        connection with a **connCloseError**.

        The resulting ``RecaptchaResponseProtocol.response`` should be equal
        to the original **responseBody**.
        """
        self.proto.dataReceived(responseBody)
        self.proto.connectionLost(failure.Failure(connCloseError()))
        self.assertEqual(responseBody, self.proto.response)
        response = self.successResultOf(self.finished)
        return response

    def test_trueResponse(self):
        """A valid API response which states 'true' should result in
        ``RecaptchaResponse.is_valid`` being ``True`` after receiving a
        ``ConnectionDone``.
        """
        responseBody = "true\nsome-reason-or-another\n"
        response = self._test(responseBody, ConnectionDone)
        self.assertIsInstance(response, txrecaptcha.RecaptchaResponse)
        self.assertTrue(response.is_valid)
        self.assertEqual(response.error_code, "some-reason-or-another")

    def test_falseResponse(self):
        """A valid API response which states 'false' should result in
        ``RecaptchaResponse.is_valid`` being ``false``.
        """
        responseBody = "false\nsome-reason-or-another\n"
        response = self._test(responseBody, ResponseDone)
        self.assertIsInstance(response, txrecaptcha.RecaptchaResponse)
        self.assertIs(response.is_valid, False)
        self.assertEqual(response.error_code, "some-reason-or-another")

    def test_responseDone(self):
        """A valid response body with a ``ResponseDone`` should result in
        ``RecaptchaResponse.is_valid`` which is ``True``.
        """
        responseBody = "true\nsome-reason-or-another\n"
        response = self._test(responseBody, ResponseDone)
        self.assertIsInstance(response, txrecaptcha.RecaptchaResponse)
        self.assertTrue(response.is_valid)
        self.assertEqual(response.error_code, "some-reason-or-another")

    def test_incompleteResponse(self):
        """ConnectionLost with an incomplete response should produce a specific
        RecaptchaResponse.error_code message.
        """
        responseBody = "true"
        response = self._test(responseBody, ConnectionLost)
        self.assertIs(response.is_valid, False)
        self.assertEqual(response.error_code,
                         "Couldn't parse response from reCaptcha API server")


class BodyProducerTests(unittest.TestCase):
    """Test for :class:`bridgedb.txrecaptcha.BodyProducer`."""

    def setUp(self):
        """Setup the tests."""
        self.content = 'Line 1\r\nLine 2\r\n'
        self.producer = txrecaptcha._BodyProducer(self.content)

    def test_interface(self):
        """BodyProducer should correctly implement IBodyProducer interface."""
        self.assertTrue(verifyObject(IBodyProducer, self.producer))

    def test_length(self):
        """BodyProducer.length should be equal to the total contect length."""
        self.assertEqual(self.producer.length, len(self.content))

    def test_body(self):
        """BodyProducer.body should be the content."""
        self.assertEqual(self.producer.body, self.content)

    def test_startProducing(self):
        """:func:`txrecaptcha.BodyProducer.startProducing` should deliver the
        original content to an IConsumer implementation.
        """
        consumer = proto_helpers.StringTransport()
        consumer.registerProducer(self.producer, False)
        self.producer.startProducing(consumer)
        self.assertEqual(consumer.value(), self.content)
        consumer.clear()


class SubmitTests(unittest.TestCase):
    """Tests for :func:`bridgedb.txrecaptcha.submit`."""

    def setUp(self):
        """Setup the tests."""
        self.challenge = (
            "03AHJ_Vutbkv3jolF5JXfJTFf5wtbdkwIJF7WA77WYjLfOUEvKW7eHBiEDKQB__7"
            "GHtUOmXC13GFYIt09HuS-ZN1j5EuDmC7bzHpHUAlpI5rbOvByypYt1vtskwnN24g"
            "zwWkrtKj8yGBWRNFljFMvtqYqHeHwJitRktSfKmV4q9VVgLBwkwlbvGUICmGaDrx"
            "dg5lYV3hpijIkmnwXygWIwoqQ0VeCgPQQ1Yw")
        self.response = "cknwnlym+ullyHLy"
        self.key = '6BdkT-18FFHAAA349auGabiqntjRJAiEM2cqPMaM8'
        self.ip = "1.2.3.4"

    def test_submit_emptyResponseField(self):
        """An empty 'recaptcha_response_field' should return a deferred which
        callbacks with a RecaptchaResponse whose error_code is
        'incorrect-captcha-sol'.
        """
        def checkResponse(response):
            """Check that the response is a
            :class:`txcaptcha.RecaptchaResponse`.
            """
            self.assertIsInstance(response, txrecaptcha.RecaptchaResponse)
            self.assertIs(response.is_valid, False)
            self.assertEqual(response.error_code, 'incorrect-captcha-sol')

        d = txrecaptcha.submit(self.challenge, '', self.key, self.ip)
        d.addCallback(checkResponse)
        return d

    def test_submit_returnsDeferred(self):
        """:func:`txrecaptcha.submit` should return a deferred."""
        response = txrecaptcha.submit(self.challenge, self.response, self.key,
                                      self.ip)
        self.assertIsInstance(response, defer.Deferred)

    def test_submit_resultIsRecaptchaResponse(self):
        """Regardless of success or failure, the deferred returned from
        :func:`txrecaptcha.submit` should be a
        :class:`txcaptcha.RecaptchaResponse`.
        """
        if not os.environ.get("CI"):
            raise SkipTest(("This test requires network so it is only run "
                            "on CI servers."))

        def checkResponse(response):
            """Check that the response is a
            :class:`txcaptcha.RecaptchaResponse`.
            """
            self.assertIsInstance(response, txrecaptcha.RecaptchaResponse)
            self.assertIsInstance(response.is_valid, bool)
            self.assertIsInstance(response.error_code, basestring)

        d = txrecaptcha.submit(self.challenge, self.response, self.key,
                               self.ip)
        d.addCallback(checkResponse)
        return d

    def tearDown(self):
        """Cleanup method for removing timed out connections on the reactor.

        This seems to be the solution for the dirty reactor due to
        ``DelayedCall``s which is mentioned at the beginning of this
        file. There doesn't seem to be any documentation anywhere which
        proposes this solution, although this seems to solve the problem.
        """
        for delay in reactor.getDelayedCalls():
            try:
                delay.cancel()
            except (AlreadyCalled, AlreadyCancelled):
                pass


class MiscTests(unittest.TestCase):
    """Tests for miscellaneous functions in :mod:`~bridgedb.txrecaptcha`."""

    def test_cbRequest(self):
        """Send a :class:`MockResponse` and check that the resulting protocol
        is a :class:`~bridgedb.txrecaptcha.RecaptchaResponseProtocol`.
        """
        response = MockResponse()
        result = txrecaptcha._cbRequest(response)
        self.assertIsInstance(result, defer.Deferred)
        self.assertIsInstance(response.protocol,
                              txrecaptcha.RecaptchaResponseProtocol)

    def test_ebRequest(self):
        """Send a :api:`twisted.python.failure.Failure` and check that the
        resulting protocol is a
        :class:`~bridgedb.txrecaptcha.RecaptchaResponseProtocol`.
        """
        msg = "Einhorn"
        fail = failure.Failure(ConnectionRefusedError(msg))
        result = txrecaptcha._ebRequest(fail)
        self.assertIsInstance(result, txrecaptcha.RecaptchaResponse)
        self.assertRegexpMatches(result.error_code, msg)

    def test_encodeIfNecessary(self):
        """:func:`txrecapcha._encodeIfNecessary` should convert unicode objects
        into strings.
        """
        origString = unicode('abc')
        self.assertIsInstance(origString, unicode)
        newString = txrecaptcha._encodeIfNecessary(origString)
        self.assertIsInstance(newString, str)



# -*- coding: utf-8 ; test-case-name: bridgedb.test.test_Main -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: please see the AUTHORS file for attributions
# :copyright: (c) 2013-2015, Isis Lovecruft
#             (c) 2013-2015, Matthew Finkel
#             (c) 2007-2015, Nick Mathewson
#             (c) 2007-2015, The Tor Project, Inc.
# :license: see LICENSE for licensing information

"""This module sets up BridgeDB and starts the servers running."""

import logging
import os
import signal
import sys
import time

from twisted.internet import reactor
from twisted.internet import task

from bridgedb import crypto
from bridgedb import persistent
from bridgedb import proxy
from bridgedb import runner
from bridgedb import util
from bridgedb.bridges import MalformedBridgeInfo
from bridgedb.bridges import MissingServerDescriptorDigest
from bridgedb.bridges import ServerDescriptorDigestMismatch
from bridgedb.bridges import ServerDescriptorWithoutNetworkstatus
from bridgedb.bridges import Bridge
from bridgedb.configure import loadConfig
from bridgedb.email.distributor import EmailDistributor
from bridgedb.https.distributor import HTTPSDistributor
from bridgedb.parse import descriptors

import bridgedb.Storage

from bridgedb import Bridges
from bridgedb.Stability import updateBridgeHistory


def load(state, hashring, clear=False):
    """Read and parse all descriptors, and load into a bridge hashring.

    Read all the appropriate bridge files from the saved
    :class:`~bridgedb.persistent.State`, parse and validate them, and then
    store them into our ``state.hashring`` instance. The ``state`` will be
    saved again at the end of this function.

    :type hashring: :class:`~bridgedb.Bridges.BridgeSplitter`
    :param hashring: A class which provides a mechanism for HMACing
        Bridges in order to assign them to hashrings.
    :param boolean clear: If True, clear all previous bridges from the
        hashring before parsing for new ones.
    """
    if not state:
        logging.fatal("bridgedb.Main.load() could not retrieve state!")
        sys.exit(2)

    if clear:
        logging.info("Clearing old bridges...")
        hashring.clear()

    logging.info("Loading bridges...")

    ignoreNetworkstatus = state.IGNORE_NETWORKSTATUS
    if ignoreNetworkstatus:
        logging.info("Ignoring BridgeAuthority networkstatus documents.")

    bridges = {}
    timestamps = {}

    logging.info("Opening networkstatus file: %s" % state.STATUS_FILE)
    networkstatuses = descriptors.parseNetworkStatusFile(state.STATUS_FILE)
    logging.debug("Closing networkstatus file: %s" % state.STATUS_FILE)

    logging.info("Processing networkstatus descriptors...")
    for router in networkstatuses:
        bridge = Bridge()
        bridge.updateFromNetworkStatus(router, ignoreNetworkstatus)
        try:
            bridge.assertOK()
        except MalformedBridgeInfo as error:
            logging.warn(str(error))
        else:
            bridges[bridge.fingerprint] = bridge

    for filename in state.BRIDGE_FILES:
        logging.info("Opening bridge-server-descriptor file: '%s'" % filename)
        serverdescriptors = descriptors.parseServerDescriptorsFile(filename)
        logging.debug("Closing bridge-server-descriptor file: '%s'" % filename)

        for router in serverdescriptors:
            try:
                bridge = bridges[router.fingerprint]
            except KeyError:
                logging.warn(
                    ("Received server descriptor for bridge '%s' which wasn't "
                     "in the networkstatus!") % router.fingerprint)
                if ignoreNetworkstatus:
                    bridge = Bridge()
                else:
                    continue

            try:
                bridge.updateFromServerDescriptor(router, ignoreNetworkstatus)
            except (ServerDescriptorWithoutNetworkstatus,
                    MissingServerDescriptorDigest,
                    ServerDescriptorDigestMismatch) as error:
                logging.warn(str(error))
                # Reject any routers whose server descriptors didn't pass
                # :meth:`~bridges.Bridge._checkServerDescriptor`, i.e. those
                # bridges who don't have corresponding networkstatus
                # documents, or whose server descriptor digests don't check
                # out:
                bridges.pop(router.fingerprint)
                continue

            if state.COLLECT_TIMESTAMPS:
                # Update timestamps from server descriptors, not from network
                # status descriptors (because networkstatus documents and
                # descriptors aren't authenticated in any way):
                if bridge.fingerprint in timestamps.keys():
                    timestamps[bridge.fingerprint].append(router.published)
                else:
                    timestamps[bridge.fingerprint] = [router.published]

    extrainfos = descriptors.parseExtraInfoFiles(*state.EXTRA_INFO_FILES)
    for fingerprint, router in extrainfos.items():
        try:
            bridges[fingerprint].updateFromExtraInfoDescriptor(router)
        except MalformedBridgeInfo as error:
            logging.warn(str(error))
        except KeyError as error:
            logging.warn(("Received extrainfo descriptor for bridge '%s', "
                          "but could not find bridge with that fingerprint.")
                         % router.fingerprint)

    inserted = 0
    logging.info("Inserting %d bridges into hashring..." % len(bridges))
    for fingerprint, bridge in bridges.items():
        # Skip insertion of bridges which are geolocated to be in one of the
        # NO_DISTRIBUTION_COUNTRIES, a.k.a. the countries we don't distribute
        # bridges from:
        if bridge.country in state.NO_DISTRIBUTION_COUNTRIES:
            logging.warn("Not distributing Bridge %s %s:%s in country %s!" %
                         (bridge, bridge.address, bridge.orPort, bridge.country))
        else:
            # If the bridge is not running, then it is skipped during the
            # insertion process.
            hashring.insert(bridge)
            inserted += 1
    logging.info("Done inserting %d bridges into hashring." % inserted)

    if state.COLLECT_TIMESTAMPS:
        reactor.callInThread(updateBridgeHistory, bridges, timestamps)

    state.save()

def _reloadFn(*args):
    """Placeholder callback function for :func:`_handleSIGHUP`."""
    return True

def _handleSIGHUP(*args):
    """Called when we receive a SIGHUP; invokes _reloadFn."""
    reactor.callInThread(_reloadFn)

def _handleSIGUSR1(*args):
    """Handler for SIGUSR1. Calls :func:`~bridgedb.runner.doDumpBridges`."""
    logging.debug("Caught SIGUSR1 signal")

    from bridgedb import runner

    logging.info("Loading saved state...")
    state = persistent.load()
    cfg = loadConfig(state.CONFIG_FILE, state.config)

    logging.info("Dumping bridge assignments to files...")
    reactor.callInThread(runner.doDumpBridges, cfg)

def replaceBridgeRings(current, replacement):
    """Replace the current thing with the new one"""
    current.hashring = replacement.hashring

def createBridgeRings(cfg, proxyList, key):
    """Create the bridge distributors defined by the config file

    :type cfg:  :class:`Conf`
    :param cfg: The current configuration, including any in-memory settings
        (i.e. settings whose values were not obtained from the config file,
        but were set via a function somewhere)
    :type proxyList: :class:`~bridgedb.proxy.ProxySet`
    :param proxyList: The container for the IP addresses of any currently
        known open proxies.
    :param bytes key: Hashring master key
    :rtype: tuple
    :returns: A BridgeSplitter hashring, an
        :class:`~bridgedb.https.distributor.HTTPSDistributor` or None, and an
        :class:`~bridgedb.email.distributor.EmailDistributor` or None.
    """
    # Create a BridgeSplitter to assign the bridges to the different
    # distributors.
    hashring = Bridges.BridgeSplitter(crypto.getHMAC(key, "Hashring-Key"))
    logging.debug("Created hashring: %r" % hashring)

    # Create ring parameters.
    ringParams = Bridges.BridgeRingParameters(needPorts=cfg.FORCE_PORTS,
                                              needFlags=cfg.FORCE_FLAGS)

    emailDistributor = ipDistributor = None
    # As appropriate, create an IP-based distributor.
    if cfg.HTTPS_DIST and cfg.HTTPS_SHARE:
        logging.debug("Setting up HTTPS Distributor...")
        ipDistributor = HTTPSDistributor(
            cfg.N_IP_CLUSTERS,
            crypto.getHMAC(key, "HTTPS-IP-Dist-Key"),
            proxyList,
            answerParameters=ringParams)
        hashring.addRing(ipDistributor.hashring, "https", cfg.HTTPS_SHARE)

    # As appropriate, create an email-based distributor.
    if cfg.EMAIL_DIST and cfg.EMAIL_SHARE:
        logging.debug("Setting up Email Distributor...")
        emailDistributor = EmailDistributor(
            crypto.getHMAC(key, "Email-Dist-Key"),
            cfg.EMAIL_DOMAIN_MAP.copy(),
            cfg.EMAIL_DOMAIN_RULES.copy(),
            answerParameters=ringParams,
            whitelist=cfg.EMAIL_WHITELIST.copy())
        hashring.addRing(emailDistributor.hashring, "email", cfg.EMAIL_SHARE)

    # As appropriate, tell the hashring to leave some bridges unallocated.
    if cfg.RESERVED_SHARE:
        hashring.addRing(Bridges.UnallocatedHolder(),
                         "unallocated",
                         cfg.RESERVED_SHARE)

    # Add pseudo distributors to hashring
    for pseudoRing in cfg.FILE_BUCKETS.keys():
        hashring.addPseudoRing(pseudoRing)

    return hashring, emailDistributor, ipDistributor

def run(options, reactor=reactor):
    """This is BridgeDB's main entry point and main runtime loop.

    Given the parsed commandline options, this function handles locating the
    configuration file, loading and parsing it, and then either (re)parsing
    plus (re)starting the servers, or dumping bridge assignments to files.

    :type options: :class:`bridgedb.parse.options.MainOptions`
    :param options: A pre-parsed options class containing any arguments and
        options given in the commandline we were called with.
    :type state: :class:`bridgedb.persistent.State`
    :ivar state: A persistent state object which holds config changes.
    :param reactor: An implementer of
        :api:`twisted.internet.interfaces.IReactorCore`. This parameter is
        mainly for testing; the default
        :api:`twisted.internet.epollreactor.EPollReactor` is fine for normal
        application runs.
    """
    # Change to the directory where we're supposed to run. This must be done
    # before parsing the config file, otherwise there will need to be two
    # copies of the config file, one in the directory BridgeDB is started in,
    # and another in the directory it changes into.
    os.chdir(options['rundir'])
    if options['verbosity'] <= 10: # Corresponds to logging.DEBUG
        print("Changed to runtime directory %r" % os.getcwd())

    config = loadConfig(options['config'])
    config.RUN_IN_DIR = options['rundir']

    # Set up logging as early as possible. We cannot import from the bridgedb
    # package any of our modules which import :mod:`logging` and start using
    # it, at least, not until :func:`safelog.configureLogging` is
    # called. Otherwise a default handler that logs to the console will be
    # created by the imported module, and all further calls to
    # :func:`logging.basicConfig` will be ignored.
    util.configureLogging(config)

    if options['dump-bridges'] or (options.subCommand is not None):
        runSubcommand(options, config)

    # Write the pidfile only after any options.subCommands are run (because
    # these exit when they are finished). Otherwise, if there is a subcommand,
    # the real PIDFILE would get overwritten with the PID of the temporary
    # bridgedb process running the subcommand.
    if config.PIDFILE:
        logging.debug("Writing server PID to file: '%s'" % config.PIDFILE)
        with open(config.PIDFILE, 'w') as pidfile:
            pidfile.write("%s\n" % os.getpid())
            pidfile.flush()

    from bridgedb import persistent

    state = persistent.State(config=config)

    from bridgedb.email.server import addServer as addSMTPServer
    from bridgedb.https.server import addWebServer

    # Load the master key, or create a new one.
    key = crypto.getKey(config.MASTER_KEY_FILE)
    proxies = proxy.ProxySet()
    emailDistributor = None
    ipDistributor = None

    # Save our state
    state.proxies = proxies
    state.key = key
    state.save()

    def reload(inThread=True):
        """Reload settings, proxy lists, and bridges.

        State should be saved before calling this method, and will be saved
        again at the end of it.

        The internal variables, ``cfg``, ``hashring``, ``proxyList``,
        ``ipDistributor``, and ``emailDistributor`` are all taken from a
        :class:`~bridgedb.persistent.State` instance, which has been saved to
        a statefile with :meth:`bridgedb.persistent.State.save`.

        :type cfg: :class:`Conf`
        :ivar cfg: The current configuration, including any in-memory
            settings (i.e. settings whose values were not obtained from the
            config file, but were set via a function somewhere)
        :type hashring: A :class:`~bridgedb.Bridges.BridgeSplitter`
        :ivar hashring: A class which takes an HMAC key and splits bridges
            into their hashring assignments.
        :type proxyList: :class:`~bridgedb.proxy.ProxySet`
        :ivar proxyList: The container for the IP addresses of any currently
            known open proxies.
        :ivar ipDistributor: A
            :class:`~bridgedb.https.distributor.HTTPSDistributor`.
        :ivar emailDistributor: A
            :class:`~bridgedb.email.distributor.EmailDistributor`.
        :ivar dict tasks: A dictionary of ``{name: task}``, where name is a
            string to associate with the ``task``, and ``task`` is some
            scheduled event, repetitive or otherwise, for the :class:`reactor
            <twisted.internet.epollreactor.EPollReactor>`. See the classes
            within the :api:`twisted.internet.tasks` module.
        """
        logging.debug("Caught SIGHUP")
        logging.info("Reloading...")

        logging.info("Loading saved state...")
        state = persistent.load()
        cfg = loadConfig(state.CONFIG_FILE, state.config)
        logging.info("Updating any changed settings...")
        state.useChangedSettings(cfg)

        level = getattr(state, 'LOGLEVEL', 'WARNING')
        logging.info("Updating log level to: '%s'" % level)
        level = getattr(logging, level)
        logging.getLogger().setLevel(level)

        logging.info("Reloading the list of open proxies...")
        for proxyfile in cfg.PROXY_LIST_FILES:
            logging.info("Loading proxies from: %s" % proxyfile)
            proxy.loadProxiesFromFile(proxyfile, state.proxies, removeStale=True)

        logging.info("Reparsing bridge descriptors...")
        (hashring,
         emailDistributorTmp,
         ipDistributorTmp) = createBridgeRings(cfg, state.proxies, key)
        logging.info("Bridges loaded: %d" % len(hashring))

        # Initialize our DB.
        bridgedb.Storage.initializeDBLock()
        bridgedb.Storage.setDBFilename(cfg.DB_FILE + ".sqlite")
        load(state, hashring, clear=False)

        if emailDistributorTmp is not None:
            emailDistributorTmp.prepopulateRings() # create default rings
            logging.info("Bridges allotted for %s distribution: %d"
                         % (emailDistributorTmp.name,
                            len(emailDistributorTmp.hashring)))
        else:
            logging.warn("No email distributor created!")

        if ipDistributorTmp is not None:
            ipDistributorTmp.prepopulateRings() # create default rings

            logging.info("Bridges allotted for %s distribution: %d"
                         % (ipDistributorTmp.name,
                            len(ipDistributorTmp.hashring)))
            logging.info("\tNum bridges:\tFilter set:")

            nSubrings  = 0
            ipSubrings = ipDistributorTmp.hashring.filterRings
            for (ringname, (filterFn, subring)) in ipSubrings.items():
                nSubrings += 1
                filterSet = ' '.join(
                    ipDistributorTmp.hashring.extractFilterNames(ringname))
                logging.info("\t%2d bridges\t%s" % (len(subring), filterSet))

            logging.info("Total subrings for %s: %d"
                         % (ipDistributorTmp.name, nSubrings))
        else:
            logging.warn("No HTTP(S) distributor created!")

        # Dump bridge pool assignments to disk.
        try:
            logging.debug("Dumping pool assignments to file: '%s'"
                          % state.ASSIGNMENTS_FILE)
            fh = open(state.ASSIGNMENTS_FILE, 'a')
            fh.write("bridge-pool-assignment %s\n" %
                     time.strftime("%Y-%m-%d %H:%M:%S"))
            hashring.dumpAssignments(fh)
            fh.flush()
            fh.close()
        except IOError:
            logging.info("I/O error while writing assignments to: '%s'"
                         % state.ASSIGNMENTS_FILE)
        state.save()

        if inThread:
            # XXX shutdown the distributors if they were previously running
            # and should now be disabled
            if ipDistributorTmp:
                reactor.callFromThread(replaceBridgeRings,
                                       ipDistributor, ipDistributorTmp)
            if emailDistributorTmp:
                reactor.callFromThread(replaceBridgeRings,
                                       emailDistributor, emailDistributorTmp)
        else:
            # We're still starting up. Return these distributors so
            # they are configured in the outer-namespace
            return emailDistributorTmp, ipDistributorTmp

    global _reloadFn
    _reloadFn = reload
    signal.signal(signal.SIGHUP, _handleSIGHUP)
    signal.signal(signal.SIGUSR1, _handleSIGUSR1)

    if reactor:
        # And actually load it to start parsing. Get back our distributors.
        emailDistributor, ipDistributor = reload(False)

        # Configure all servers:
        if config.HTTPS_DIST and config.HTTPS_SHARE:
            addWebServer(config, ipDistributor)
        if config.EMAIL_DIST and config.EMAIL_SHARE:
            addSMTPServer(config, emailDistributor)

        tasks = {}

        # Setup all our repeating tasks:
        if config.TASKS['GET_TOR_EXIT_LIST']:
            tasks['GET_TOR_EXIT_LIST'] = task.LoopingCall(
                proxy.downloadTorExits,
                state.proxies,
                config.SERVER_PUBLIC_EXTERNAL_IP)

        if config.TASKS.get('DELETE_UNPARSEABLE_DESCRIPTORS'):
            delUnparseableSecs = config.TASKS['DELETE_UNPARSEABLE_DESCRIPTORS']
        else:
            delUnparseableSecs = 24 * 60 * 60  # Default to 24 hours

        # We use the directory name of STATUS_FILE, since that directory
        # is where the *.unparseable descriptor files will be written to.
        tasks['DELETE_UNPARSEABLE_DESCRIPTORS'] = task.LoopingCall(
            runner.cleanupUnparseableDescriptors,
            os.path.dirname(config.STATUS_FILE), delUnparseableSecs)

        # Schedule all configured repeating tasks:
        for name, seconds in config.TASKS.items():
            if seconds:
                try:
                    tasks[name].start(abs(seconds))
                except KeyError:
                    logging.info("Task %s is disabled and will not run." % name)
                else:
                    logging.info("Scheduled task %s to run every %s seconds."
                                 % (name, seconds))

    # Actually run the servers.
    try:
        if reactor and not reactor.running:
            logging.info("Starting reactors.")
            reactor.run()
    except KeyboardInterrupt:
        logging.fatal("Received keyboard interrupt. Shutting down...")
    finally:
        if config.PIDFILE:
            os.unlink(config.PIDFILE)
        logging.info("Exiting...")
        sys.exit()

def runSubcommand(options, config):
    """Run a subcommand from the 'Commands' section of the bridgedb help menu.

    :type options: :class:`bridgedb.opt.MainOptions`
    :param options: A pre-parsed options class containing any arguments and
        options given in the commandline we were called with.
    :type config: :class:`bridgedb.Main.Conf`
    :param config: The current configuration.
    :raises: :exc:`SystemExit` when all subCommands and subOptions have
        finished running.
    """
    # Make sure that the runner module is only imported after logging is set
    # up, otherwise we run into the same logging configuration problem as
    # mentioned above with the email.server and https.server.
    from bridgedb import runner

    statuscode = 0

    if options.subCommand is not None:
        logging.debug("Running BridgeDB command: '%s'" % options.subCommand)

        if 'descriptors' in options.subOptions:
            statuscode = runner.generateDescriptors(
                options.subOptions['descriptors'], config.RUN_IN_DIR)

        logging.info("Subcommand '%s' finished with status %s."
                     % (options.subCommand, statuscode))
        sys.exit(statuscode)



# -*- coding: utf-8 -*-
#
# This file is part of BridgeDB, a Tor bridge distribution system.
#
# :authors: Isis Lovecruft 0xA3ADB67A2CDB8B35 <isis@torproject.org>
# :copyright: (c) 2013-2015 Isis Lovecruft
#             (c) 2007-2015, The Tor Project, Inc.
# :license: see included LICENSE for information

"""Tests for :mod:`bridgedb.email.distributor`."""

from __future__ import print_function

import logging
import tempfile
import os

from twisted.internet.task import Clock
from twisted.trial import unittest

import bridgedb.Storage

from bridgedb.bridges import Bridge
from bridgedb.email.distributor import EmailDistributor
from bridgedb.email.distributor import IgnoreEmail
from bridgedb.email.distributor import TooSoonEmail
from bridgedb.email.request import EmailBridgeRequest
from bridgedb.parse.addr import BadEmail
from bridgedb.parse.addr import UnsupportedDomain
from bridgedb.parse.addr import normalizeEmail

from .util import generateFakeBridges

logging.disable(50)


BRIDGES = generateFakeBridges()


class EmailDistributorTests(unittest.TestCase):
    """Tests for :class:`bridgedb.email.distributor.EmailDistributor`."""

    # Fail any tests which take longer than 15 seconds.
    timeout = 15

    def setUp(self):
        self.fd, self.fname = tempfile.mkstemp(suffix=".sqlite", dir=os.getcwd())
        bridgedb.Storage.initializeDBLock()
        self.db = bridgedb.Storage.openDatabase(self.fname)
        bridgedb.Storage.setDBFilename(self.fname)

        self.bridges = BRIDGES
        self.key = 'aQpeOFIj8q20s98awfoiq23rpOIjFaqpEWFoij1X'
        self.domainmap = {
            'example.com':      'example.com',
            'dkim.example.com': 'dkim.example.com',
        }
        self.domainrules = {
            'example.com':      ['ignore_dots'],
            'dkim.example.com': ['dkim', 'ignore_dots']
        }

    def tearDown(self):
        self.db.close()
        os.close(self.fd)
        os.unlink(self.fname)

    def makeClientRequest(self, clientEmailAddress):
        bridgeRequest = EmailBridgeRequest()
        bridgeRequest.client = clientEmailAddress
        bridgeRequest.isValid(True)
        bridgeRequest.generateFilters()
        return bridgeRequest

    def test_EmailDistributor_getBridges_default_client(self):
        """If EmailBridgeRequest.client was not set, then getBridges() should
        raise a bridgedb.parse.addr.BadEmail exception.
        """
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules)
        [dist.hashring.insert(bridge) for bridge in self.bridges]

        # The "default" client is literally the string "default", see
        # bridgedb.bridgerequest.BridgeRequestBase.
        bridgeRequest = self.makeClientRequest('default')

        self.assertRaises(BadEmail, dist.getBridges, bridgeRequest, 1)

    def test_EmailDistributor_getBridges_with_whitelist(self):
        """If an email address is in the whitelist, it should get a response
        every time it asks (i.e. no rate-limiting).
        """
        # The whitelist should be in the form {EMAIL: GPG_FINGERPRINT}
        whitelist = {'white@list.ed': '0123456789ABCDEF0123456789ABCDEF01234567'}
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules,
                                whitelist=whitelist)
        [dist.hashring.insert(bridge) for bridge in self.bridges]

        # A request from a whitelisted address should always get a response.
        bridgeRequest = self.makeClientRequest('white@list.ed')
        for i in range(5):
            bridges = dist.getBridges(bridgeRequest, 1)
            self.assertEqual(len(bridges), 3)

    def test_EmailDistributor_getBridges_rate_limit_multiple_clients(self):
        """Each client should be rate-limited separately."""
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules)
        [dist.hashring.insert(bridge) for bridge in self.bridges]

        bridgeRequest1 = self.makeClientRequest('abc@example.com')
        bridgeRequest2 = self.makeClientRequest('def@example.com')
        bridgeRequest3 = self.makeClientRequest('ghi@example.com')

        # The first request from 'abc' should get a response with bridges.
        self.assertEqual(len(dist.getBridges(bridgeRequest1, 1)), 3)
        # The second from 'abc' gets a warning.
        self.assertRaises(TooSoonEmail, dist.getBridges, bridgeRequest1, 1)
        # The first request from 'def' should get a response with bridges.
        self.assertEqual(len(dist.getBridges(bridgeRequest2, 1)), 3)
        # The third from 'abc' is ignored.
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest1, 1)
        # The second from 'def' gets a warning.
        self.assertRaises(TooSoonEmail, dist.getBridges, bridgeRequest2, 1)
        # The third from 'def' is ignored.
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest2, 1)
        # The fourth from 'abc' is ignored.
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest1, 1)
        # The first request from 'ghi' should get a response with bridges.
        self.assertEqual(len(dist.getBridges(bridgeRequest3, 1)), 3)
        # The second from 'ghi' gets a warning.
        self.assertRaises(TooSoonEmail, dist.getBridges, bridgeRequest3, 1)
        # The third from 'ghi' is ignored.
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest3, 1)
        # The fourth from 'ghi' is ignored.
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest3, 1)

    def test_EmailDistributor_getBridges_rate_limit(self):
        """A client's first email should return bridges.  The second should
        return a warning, and the third should receive no response.
        """
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules)
        [dist.hashring.insert(bridge) for bridge in self.bridges]

        bridgeRequest = self.makeClientRequest('abc@example.com')

        # The first request should get a response with bridges.
        bridges = dist.getBridges(bridgeRequest, 1)
        self.assertGreater(len(bridges), 0)
        [self.assertIsInstance(b, Bridge) for b in bridges]
        self.assertEqual(len(bridges), 3)

        # The second gets a warning, and the third is ignored.
        self.assertRaises(TooSoonEmail, dist.getBridges, bridgeRequest, 1)
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest, 1)

    def test_EmailDistributor_getBridges_rate_limit_expiry(self):
        """A client's first email should return bridges.  The second should
        return a warning, and the third should receive no response.  After the
        EmailDistributor.emailRateMax is up, the client should be able to
        receive a response again.
        """
        clock = Clock()
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules)
        [dist.hashring.insert(bridge) for bridge in self.bridges]

        bridgeRequest = self.makeClientRequest('abc@example.com')

        # The first request should get a response with bridges.
        self.assertEqual(len(dist.getBridges(bridgeRequest, 1, clock)), 3)
        # The second gets a warning, and the rest are ignored.
        self.assertRaises(TooSoonEmail, dist.getBridges, bridgeRequest, 1, clock)
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest, 1, clock)
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest, 1, clock)
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest, 1, clock)

        clock.advance(2 * dist.emailRateMax)

        # The client should again a response with bridges.
        self.assertEqual(len(dist.getBridges(bridgeRequest, 1)), 3, clock)

    def test_EmailDistributor_cleanDatabase(self):
        """Calling cleanDatabase() should cleanup email times in database, but
        not allow clients who have been recently warned and/or ignored to
        receive a response again until the remainder of their MAX_EMAIL_RATE
        time is up.
        """
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules)
        [dist.hashring.insert(bridge) for bridge in self.bridges]

        bridgeRequest = self.makeClientRequest('abc@example.com')

        # The first request should get a response with bridges.
        self.assertEqual(len(dist.getBridges(bridgeRequest, 1)), 3)
        # The second gets a warning, and the third is ignored.
        self.assertRaises(TooSoonEmail, dist.getBridges, bridgeRequest, 1)
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest, 1)

        dist.cleanDatabase()

        # Cleaning the warning email times in the database shouldn't cause
        # 'abc@example.com' to be able to email again, because only the times
        # which aren't older than EMAIL_MAX_RATE should be cleared.
        self.assertRaises(IgnoreEmail,  dist.getBridges, bridgeRequest, 1)

    def test_EmailDistributor_prepopulateRings(self):
        """Calling prepopulateRings() should add two rings to the
        EmailDistributor.hashring.
        """
        dist = EmailDistributor(self.key, self.domainmap, self.domainrules)

        # There shouldn't be any subrings yet.
        self.assertEqual(len(dist.hashring.filterRings), 0)

        dist.prepopulateRings()

        # There should now be two subrings, but the subrings should be empty.
        self.assertEqual(len(dist.hashring.filterRings), 2)
        for (filtre, subring) in dist.hashring.filterRings.values():
            self.assertEqual(len(subring), 0)

        # The subrings in this Distributor have gross names, because the
        # filter functions (including their addresses in memory!) are used as
        # the subring names.  In this case, we should have something like:
        #
        #     frozenset([<function byIPv6 at 0x7eff7ad7fc80>])
        #
        # and
        #
        #     frozenset([<function byIPv4 at 0x7eff7ad7fc08>])
        #
        # So we have to join the strings together and check the whole thing,
        # since we have no other way to use these stupid subring names to
        # index into the dictionary they are stored in, because the memory
        # addresses are unknowable until runtime.

        # There should be an IPv4 subring and an IPv6 ring:
        ringnames = dist.hashring.filterRings.keys()
        self.failUnlessIn("IPv4", "".join([str(ringname) for ringname in ringnames]))
        self.failUnlessIn("IPv6", "".join([str(ringname) for ringname in ringnames]))

        [dist.hashring.insert(bridge) for bridge in self.bridges]

        # There should still be two subrings.
        self.assertEqual(len(dist.hashring.filterRings), 2)
        for (filtre, subring) in dist.hashring.filterRings.values():
            self.assertGreater(len(subring), 0)

        # Ugh, the hashring code is so gross looking.
        subrings = dist.hashring.filterRings
        subring1 = subrings.values()[0][1]
        subring2 = subrings.values()[1][1]
        # Each subring should have roughly the same number of bridges.
        # (Having ±10 bridges in either ring, out of 500 bridges total, should
        # be so bad.)
        self.assertApproximates(len(subring1), len(subring2), 10)

    def test_EmailDistributor_unsupported_domain(self):
        """An unsupported domain should raise an UnsupportedDomain exception."""
        self.assertRaises(UnsupportedDomain, normalizeEmail,
                          'bad@email.com', self.domainmap, self.domainrules)
